C      PROGRAM XSAARMA.FOR(SIMULATED ANNEALING SUBSET ARMA- MONTE CARLO)
C      DRIVER ROUTINE HYBRID ADAPTED FOR SUBSET ARMA MODEL ESTIMATION
C      INSERTED INTO A MONTE CARLO LOOP FOR SIMULATION AND EVALUATION
C      ARMA MODEL IS ASSUMED WITH MAX AR, MA LAG. THE PARAMETER SPACE IS
C      SEARCHED BY SIMULATED ANNEALING (Algorithm AS298, ROUTINE HYBRID)
C      OBJECTIVE FUNCTION TO BE MINIMIZED IS THE SUM OF SQUARES FUNCTION
C      EVALUATION IS DONE BY ROUTINE CHINIT (QUICK ARMA SSQ COMPUTATION)
C      THE TRADITIONAL MINIMIZATION ROUTINE IS THE LEVENBERG-MARQUARDT
C      ALGORITHM GLM1, EMBEDDED INTO THE ALL-INCLUSIVE ROUTINE ESTIMA
C
C      Brooks,S.P. (1995) A hybrid optimization algorithm
C      (Algorithm AS 298), Applied Statistics, 44, 530-552.
C
C      Marquardt, D. W., 1963. An algorithm for least squares estimation
C      of nonlinear parameters. Journal of the Society of Industrial and
C      Applied Mathematics, 11, 431--441.

      IMPLICIT DOUBLE PRECISION(A-H,O-Z)
      CHARACTER*72  PNAME
      REAL*8  LOW
      COMPLEX  XCOEFF, XROOTS
      PARAMETER(MAXP=20,MAXQ=20,MAXN=5500,MSCART=10500,MTOT=MAXN+MSCART
     +, MX=MAXP+MAXQ+2, MXX= MAXP+MAXQ, MXX1=MXX+1, MXTEMP=501
     +, MAXNAC=36, MAXNC1=MAXNAC+1
     +, LIW=10000 ,LW=10000)

C      TIME SERIES AND MODEL COEFFICIENTS : arrays dimension

      DIMENSION P(MAXP), Q(MAXQ), W(MAXN), E(MAXN), ZHZ(MTOT), RES(MTOT) 
     +, RHO(MAXNC1), COV(MAXNC1), XCOEFF(MXX1), XROOTS(MXX)

C      ROUTINE HYBRID FOR SIMULATED ANNEALING : arrays dimension 

      DIMENSION BL(MXX), BU(MXX), LOW(MXX1), PARAM(MXX)
     +, POINTS(MXX, MXTEMP)

C      ROUTINE GLM1 FOR ARMA(k, l) ESTIMATING : arrays dimension

      DIMENSION  hdifz(mx), hsigns(mx), hq(mx), hp(mx), he(mx), hphi(mx)
     +, HTB(Mx), hdelz(lw), ha(lw), hd(lw), TH(MX)
     +, IW(LIW), WI(LW)                      

C      Monte Carlo results : arrays dimension

      DIMENSION indep(maxp),indeq(maxq)
     +,wp(maxp),wp2(maxp),wq(maxq),wq2(maxq)
     +, WPAR(MXX), WPAR2(MXX), WNUM(MXX)

      COMMON/UNIIN/MFAC(3),AVEPA,ND(3),IOD(3),INC(6)                    
     *,            IOPA(10),UPA(10),ITREND ,TREPA                       
      COMMON/UNIOUT/NAR,NMA,NDIFAC,NARMAF,MAX1,MAX11,MAX2,MBO           
      COMMON/TSOBS/IWBF,NOB,X(5500)                                     
        common /rand/  ix, iy,  iz

      EXTERNAL  FUNCT, LSQFUN

      DATA EPS1 , EPS2 /1.5D-5 , 1.5D-5/

C     THE INPUT TIME SERIES 'W' HAS 'N' OBSERVATIONS.                
C     'W' IS COPIED IN 'X' AND 'N' IN 'NOB' FOR USE IN COMMON /TSOBS/
C     THE COMMON STATEMENTS 'UNIIN' AND 'UNIOUT' SPECIFY THE arima MODEL 
C     MFAC(1),MFAC(2),MFAC(3) ARE THE NUMBERS OF FACTORS OF THE ar PART,
C     INTEGRATED PART AND ma PART. NDIFAC IS SET TO 0 (NO DIFFERENCING)
C     THE AVERAGE OF THE TIME SERIES IS ALWAYS ESTIMATED AS NDIFAC = 0            
C     max=MFAC(1)+MFAC(3) IS THE NUMBER OF ar AND ma FACTORS            
C     FOLLOWING THE USUAL SEQUENCE, INC(1) IS THE NUMBER OF PARAMETERS  
C     OF THE FIRST FACTOR, INC(2) THAT OF THE SECOND FACTOR, INC(3) THAT
C     OF THE THIRD ONE, AND SO ON. THE SAME ORDER MUST OBEY THE ORDER   
C     PARAMETERS AND INITIAL GUESSES, IOPA AND UPA RESPECTIVELY
C     FINALLY, IF A CONSTANT 'TREPA' MUST BE ADDED TO MODEL, ITREND<>0 

      OPEN(UNIT=5, FILE='xsaarma.txt')
      READ(5, *) MP, MQ, N, NSCART, NAC, MAXIT
      READ(5, *) (P(I), I=1,  MP)
      READ(5, *) (Q(I)  , I=1 ,  MQ)
      READ(5,*)ITREND                                                   
      IF(ITREND .NE. 0)READ(5,*)TREPA                                   
      READ(5 , *)  NTEMP , T0 , RRHO , NT
      READ(5 , *) NPARAM
      READ(5 , *) (BL(J) , J =1  ,   NPARAM)
      READ(5 , *) (BU(J) , J =1  ,   NPARAM)
      READ(5, *)  PSBST
      READ(5, * ) MCIT, IJX,IJY,IJZ,VARE
      READ(5,100)  PNAME
100   FORMAT(A72)
      CLOSE(UNIT=5)
      
      CALL CHTRL(P,MP,ICONTR,XCOEFF,XROOTS)
      if(icontr. lt. 0)  stop  1
      CALL CHTRL(Q,MQ,ICONTR,XCOEFF,XROOTS)
      if(icontr. lt. 0)  stop  1
      call  xsubst(p, mp, q, mq, indep, indeq, npzp, nqzq)

      OPEN(UNIT=6, FILE= PNAME, status='unknown', access='append')
      WRITE(6, 601)
601   FORMAT(1H  ,'XSAARMA.FOR: DRIVER ROUTINE HYBRID (CHINIT+GLM1)')
      WRITE(6, 102)  MP, MQ, N, NSCART, NAC, MAXIT
102   FORMAT(1H  ,'MP,MQ,N,NSCART,NAC,MAXIT = ',  6I5)
      WRITE(6, 103)
103   FORMAT(1H  ,'AR PARAMETERS P')
      WRITE(6, 104) (P(I), I=1,  MP)
104   FORMAT(1H  ,7F10.4)
      WRITE(6, 105)
105   FORMAT(1H  ,'MA PARAMETERS Q')
      WRITE(6, 106) (Q(I), I=1,  MQ)
106   FORMAT(1H  ,7F10.4)
      write(6, 108)  npzp
108   format(1h   ,'numero di parametri autoregressivi npzp ', I5)
      write(6, 109) (indep(i), i =1 , npzp)
109   format(1h   ,'lag:',  22I3) 
      write(6, 110)  nqzq
110   format(1h   ,'numero di parametri somma mobile nqzq ', I5)
      write(6, 111) (indeq(i), i =1 , nqzq)
111   format(1h   ,'lag:',  22I3) 
      WRITE(6, 107)  VARE
107   FORMAT(1H   ,'ERROR VARIANCE VARE = ', D14.6)
	 WRITE(6 ,1061) nparam
1061    FORMAT(1H  ,'Numero di parametri da stimare nparam = ',I4)
       write(6 , 220)
220    format(1h  , 'parameters lower bound')
       write(6 ,  223)(BL(j) , j =1  ,   nparam)
223    format(7f10.4)
       write(6 , 221)
221    format(1h  , 'parameters upper bound')
       write(6 ,  203)(BU(j) , j =1  ,   nparam)
203    format(7f10.4)
       write(6 , 225) psbst
225    format(1h   ,'prob that any coeff is non zero psbst = ', F7.4)
       write(6 , 222)
222    format(1h  , 'simulated annealing parameters t0, n, rho, nt')
       write(6 ,  224) t0 , ntemp , rrho , nt
224    format(1h   , F10.4, I10, F10.4, I10)
       WRITE(6, 206) MCIT
206    FORMAT(1H  ,'N. REPL. MCIT = ',  I10)
       write(6 , 204)  
204    format(1h , 'data printout file')
       write(6 , 205)      pname
205    format(1h  ,  A72)
      WRITE(6, 610)
610   FORMAT(1H  ,'//')

C     THE ROUTINE ESTIMA USES THE LEVENBERG - MARQUARDT ALGORITHM GLM1
C     FOR PARAMETER ESTIMATION: PARAMETERS ARE TH(1) = AVEPA, TH(2)...
C     TH(MP + 1) = AR COEFF.S, TH(MP + 2)...TH(MP+MQ+1) = MA COEFF.S,
C     TH(MP + MQ + 2) = TH(NX) = TREPA (NON ZERO ONLY IF ITREND .NE. 0)

      NOB  =  N
      NHB  =  N
      NTOT =  N +  NSCART
      NAC1  =  NAC  + 1
      NX =  MP + MQ + 2
      MIT  =  MAXIT

      ix = IJX
      iy = IJY
      iz = IJZ

      KITER=0
      IIDEN = 0
      DO  401  I =1  ,  NPZP
      WP(I)  =  0D0
401   WP2(I) = 0D0
      DO  402  I =1  ,  NQZQ
      WQ(I) =  0D0
402   WQ2(I) =  0D0
      WVART = 0D0
      WVART2 =  0D0
      WVAR = 0D0
      WVAR2  = 0D0
      WREST =  0D0
      WREST2 = 0D0
      WRES =  0D0
      WRES2  =  0D0
      WAICT =0D0
      WAICT2 = 0D0
      WAIC  =  0D0
      WAIC2 =  0D0
      WINDP1  =  0
      WINDQ1  =  0
      WINDP  =  0
      WINDP2  =  0
      WINDQ  =  0
      WINDQ2  =  0
      DO  403  I =1 , NPARAM
      WPAR(I)  = 0D0
      WPAR2(I)  =  0D0
403   WNUM(I)  =  0D0

      DO  501  IREPL = 1 ,  MCIT

      write(*,*)  irepl

      CALL SIMTRU(W,N,NSCART,NTOT,ZHZ,RES,   P,MP,Q,MQ,VARE)

c      write(6, 20209) (w(iiii) , iiii =  1 ,  n)
c20209 format(1h ,   10f7.4)      

C     MEAN - DELETED TIME SERIES
c
c      CALL  SMEADL(W, NHB, DM)

      nparam1 = nparam + 1

      CALL HYBRID(NPARAM,PARAM,NTEMP,T0,RRHO,NT,BL,BU,POINTS,LOW,IFAULT
     +,NPARAM1,NX,NAC,NAC1,W,DM,RHO,COV,P,MP,Q,MQ,E,TH,NP
     +,HDIFZ,HSIGNS,HQ,HP,HE,HPHI,HTB,HDELZ,HA,HD,EPS1,EPS2,MIT,NIT,NF
     +,ITER2,SSQ,FUNCT,IEXIT,NHB,MXX,PSBST,XCOEFF,XROOTS,AIC
     +,IW,LIW,WI,LW)

c      WRITE(6, 231) IFAULT, IEXIT
c231   FORMAT(1H  ,'Exit Hybrid ifault, iexit = ', 2I8)
c      WRITE(6, 230) (LOW(I), I=1, NPARAM1)
c230   FORMAT(1H   ,7F10.4)
      VHV  =  COV(1)
      DSD  =  DSQRT(DABS(VHV))
c      write(6, 237) DM, VHV, DSD 
c237   format(1h  ,'Input data mean,var,sqm =', 3D12.4)
c      write(6, 236) VHV * DBLE(N) 
c236   format(1h  ,'Deviance =', D12.4)
c      write(6, 238) 
c238   format(1h ,'Autocorrelation function')
c      write(6, 239) (rho(j), j=1, nac1)
c239    format(1h  , 7f10.4)
c       write(6 , 232)  NP , NOB ,IWBF, MIT
c232    format(1h  ,  'no. param n = ', i5,5x, 'no. obs nob = ' , i10
c     +/   1h  ,'BACKFORECASTING FLAG IWBF = ',i5
c     +/ 1h  ,'max no. iter. Levenberg-Marquardt algorithm mit = ',i5)
c      WRITE(6, 250) NIT , NF
c250   FORMAT(1H  ,'ITERATIONS NIT = ', I5, 5X,'LSQFUN CALLS NF = ', I5)
c      WRITE(6 , 234)	SSQ
c234   FORMAT(1H  ,'MINIMUM SUM OF SQUARES SSQ = ', D20.10)
c      WRITE(6, 235) NP
c235   FORMAT(1H  ,'PARAMETERS VECTOR ESTIMATES ', I5)
c      WRITE(6, 240) (TH(J), J=1,NP)
c240    format(7f10.4)
c      IF(NDIFAC.LE.0) WRITE(6 , 241) AVEPA
c241   FORMAT(1H  ,'AVEPA = ', D20.10)
c      WRITE(6, 255) 
c255   FORMAT(1H  ,'Numero di parameteri AR e MA')
      kkmax  =  mfac(1) + mfac(3)
      NARMAP=0                                                          
      DO 430 J=1,kkMAX                                                  
c      write(6, 242)  inc(j)
430   NARMAP=NARMAP+INC(J)                                              
c      write(6,242)(IOPA(J),J=1,NARMAP)                                  
c242   format(3I20)
c      write(6,243)(UPA(J),J=1,NARMAP)                                   
c243   format(3D20.10)
c      IF(ITREND .NE. 0)write(6, 244)TREPA                               
c244    format(1h  ,  'TREPA  =  ', D20.10)
      aic  =  dble(n) * dlog(dabs(ssq / dble(n))) + 2d0 * dble(narmap)
c      write(6, 646)  aic, ssq, n, narmap
c646   format(1h   ,'aic,ssq,n,narmap ', 2D14.6, 2I8)
c      WRITE(6 , 245)
c245   FORMAT(1H  , 'PARAMETERS STANDARD ERRORS')
c      WRITE(6, 246) (HE(J), J= 1, NP)      
c246   format(1h  ,  10f7.4)
c      WRITE(6 , 247)
c247   FORMAT(1H  , 'PARAMETERS CORRELATION MATRIX') 
c      ICOL2 = 0
c      DO  413  I = 1 ,  NP
c      ICOL1 = ICOL2+1
c      ICOL2 = ICOL2+NP
c413   WRITE(6, 246) (HA(J), J=ICOL1, ICOL2)      
c      WRITE(6 , 248)
c248   FORMAT(1H  , 'PARAMETERS VARIANCE-COVARIANCE MATRIX') 
c      ICOL2 = 0
c      DO  412  I = 1 ,  NP
c      ICOL1 = ICOL2+1
c      ICOL2 = ICOL2+NP
c412   WRITE(6, 246) (HD(J), J=ICOL1, ICOL2)      
c      WRITE(6, 249)
249   FORMAT(1H  ,'//')

C       MONTE CARLO STATISTICS : all replications

      KITER  =  KITER + 1
      WVART = WVART +  VHV
      WVART2=  WVART2 + VHV ** 2
      WREST = WREST +  SSQ / DBLE(N)
      WREST2=  WREST2 + (SSQ / DBLE(N)) ** 2
      WAICT = WAICT +  AIC
      WAICT2=  WAICT2 + AIC ** 2
      WINDP = WINDP + INC(1)
      WINDP2 = WINDP2 + INC(1) * INC(1) 
      WINDQ  =  WINDQ + INC(2)
      WINDQ2  =  WINDQ2 + INC(2) * INC(2)
      n1yp =  inc(1)
      if(n1yp  .gt.  0)  then
      DO  259 I = 1 , n1yp
      II =  IOPA(I)
      if(ii.gt.0)  then
      WNUM(II) = WNUM(II) + 1D0
      WPAR(II)  =  WPAR(II) + UPA(I)
      WPAR2(II)  =   WPAR2(II) +  UPA(I) ** 2
      endif
259   continue
      endif
      n1yp1 =  inc(1) + 1
      n1yq  =   narmap
      if(n1yq  .ge.  n1yp1)  then
      DO  269 I = n1yp1 , n1yq
      II =  mp + IOPA(I)
      if(ii.gt.0)  then
      WNUM(II) = WNUM(II) + 1D0
      WPAR(II)  =  WPAR(II) + UPA(I)
      WPAR2(II)  =   WPAR2(II) +  UPA(I) ** 2
      endif
269   continue
      endif

C       MONTE CARLO STATISTICS : exact identifications only

      nnp1  =  mfac(1)
      nnpn = inc(nnp1)
      nnq1  =  mfac(1) + mfac(3)
      nnqn   =   inc(nnq1)
      if(nnpn  .eq.  npzp)  then
      WINDP1 = WINDP1 + 1
      endif
      if(nnqn .eq. nqzq)  then
      WINDQ1  =  WINDQ1 + 1
      endif
C      if(nnpn  .eq.  npzp .and.  nnqn .eq. nqzq)  then
            ihflag  = 1
C            jj  =  0
C           if(nnpn  .gt.  0)  then
C           do  661 j =  1,  nnpn
C            jj =  jj + 1
C            if(iopa(jj) .ne. indep(j)) then
C              ihflag  =  0
C              endif
C661        continue               
C           endif
      JJ1 =  1
      JJ2  =  NNPN
       DO 661  J = 1 , NPZP
       IHOPA= INDEP(J)
        IRK = 0
       DO 1662 JJ = JJ1 , JJ2
       IF(IHOPA  .EQ.  IOPA(JJ)) THEN
        IRK = 1
        GOTO 1666
        ENDIF
1662   CONTINUE
1666   CONTINUE
      IF(IRK .EQ.  0) THEN
           IHFLAG = 0
           GOTO  1701
           ENDIF
661    CONTINUE
C           if(nnqn  .gt.  0)  then
C           do  662 j =  1,  nnqn
C            jj =  jj + 1
C            if(iopa(jj) .ne. indeq(j)) then
C              ihflag  =  0
C              endif
C662        continue               
C           endif
      JJ1 =  NNPN + 1
      JJ2  =  NNPN +  NNQN
       DO 662  J = 1 , NQZQ
       IHOPA= INDEQ(J)
        IRK = 0
       DO 1664 JJ = JJ1 , JJ2
       IF(IHOPA  .EQ.  IOPA(JJ)) THEN
        IRK = 1
        GOTO 1668
        ENDIF
1664   CONTINUE
1668   CONTINUE
      IF(IRK .EQ.  0) THEN
           IHFLAG = 0
           GOTO  1701
           ENDIF
662    CONTINUE
1701      CONTINUE
       if(ihflag  .eq.  1)  then
           IIDEN = IIDEN + 1
            jj  =  0
           if(nnpn  .gt.  0)  then
           do  663 j =  1,  nnpn
            jj =  jj + 1
           wp(j) =  wp(j) + upa(jj)
           wp2(j) =  wp2(j) + upa(jj) * upa(jj)
663        continue               
           endif
           if(nnqn  .gt.  0)  then
           do  664 j =  1,  nnqn
            jj =  jj + 1
           wq(j) =  wq(j) + upa(jj)
           wq2(j) =  wq2(j) + upa(jj) * upa(jj)
664        continue               
           endif
      WVAR = WVAR +  VHV
      WVAR2=  WVAR2 + VHV ** 2
      WRES = WRES +  SSQ / DBLE(N)
      WRES2=  WRES2 + (SSQ / DBLE(N)) ** 2
      WAIC = WAIC +  AIC
      WAIC2=  WAIC2 + AIC ** 2
       endif
C          endif

501   CONTINUE

C       SUMMARY  MONTE CARIO STATISTICS

      IF(IIDEN  .GT. 0)  THEN
      DO  631  I =1  ,  NPZP
      WP(I)  =  WP(I) / DBLE(IIDEN)
631   WP2(I) = DSQRT(DABS(WP2(I) / DBLE(IIDEN) -  WP(I) * WP(I)))
      DO  632  I =1  ,  NQZQ
      WQ(I)  =  WQ(I) / DBLE(IIDEN)
632   WQ2(I) = DSQRT(DABS(WQ2(I) / DBLE(IIDEN) -  WQ(I) * WQ(I)))
      WVAR = WVAR / DBLE(IIDEN)
      WVAR2  = DSQRT(DABS(WVAR2 / DBLE(IIDEN) -  WVAR * WVAR))
      WRES =  WRES / DBLE(IIDEN)
      WRES2  =  DSQRT(DABS(WRES2 / DBLE(IIDEN) -  WRES * WRES))
      WAIC  =  WAIC / DBLE(IIDEN)
      WAIC2 =  DSQRT(DABS(WAIC2 / DBLE(IIDEN) -  WAIC * WAIC))
      ENDIF

      IF(KITER .GT. 0)  THEN
      DO  635  I =1  ,  NPARAM
      IF(WNUM(I) .GT. 0D0) THEN
      WPAR(I)  =  WPAR(I) / WNUM(I)
      WPAR2(I) = DSQRT(DABS(WPAR2(I) / WNUM(I) - WPAR(I) * WPAR(I)))
      ENDIF
635   CONTINUE
      WINDP = WINDP / DBLE(KITER )
      WINDP2  = DSQRT(DABS(WINDP2 / DBLE(KITER ) -  WINDP * WINDP))
      WINDQ = WINDQ / DBLE(KITER )
      WINDQ2  = DSQRT(DABS(WINDQ2 / DBLE(KITER ) -  WINDQ * WINDQ))
      WVART = WVART / DBLE(KITER )
      WVART2  = DSQRT(DABS(WVART2 / DBLE(KITER ) -  WVART * WVART))
      WREST =  WREST / DBLE(KITER )
      WREST2  =  DSQRT(DABS(WREST2 / DBLE(KITER ) -  WREST * WREST))
      WAICT  =  WAICT / DBLE(KITER )
      WAICT2 =  DSQRT(DABS(WAICT2 / DBLE(KITER ) -  WAICT * WAICT))
      ENDIF


      WRITE(6, 60100) IIDEN, KITER
60100 FORMAT(1H ,'Monte Carlo statistics:iiden= ',I5,5X, 'kiter = ', I5)
      write(6, 60200)
60200 format(1h   ,'Id. esatta: parametri (media, sqm)')
      DO  603  I =1  ,  NPZP
603   write(6, 40100) wp(i), wp2(i)
      DO  604  I =1  ,  NQZQ
604   write(6, 40100) wq(i), wq2(i)
      write(6, 60600)
60600 format(1h   ,'Id. esatta: var. serie, res., aic (media, sqm)')
      write(6,40100) WVAR , WVAR2 
      write(6,40100) Wres , Wres2 
      write(6,40100) Waic , Waic2 
      write(6, 60400) WINDP1
60400 format(1h   ,'numero di id. numero par. AR ', F10.4)
      write(6, 60500) WINDQ1
60500 format(1h   ,'numero di id. numero par. MA ', F10.4)
      write(6, 60700)
60700 format(1h   ,'Tutte le repl.: num. par. AR e MA (media, sqm)')
      write(6,40100) Windp , Windp2 
      write(6,40100) Windq , Windq2 
      write(6, 60800)
60800 format(1h   ,'Tutte le repl.: parametri (media, sqm)')
      DO  609  I =1  ,  NPARAM
609   write(6, 40200) wpar(i), wpar2(i), WNUM(I)
40200 FORMAT(1H   , 3F10.4)
      write(6, 60300)
60300 format(1h   ,'Tutte le repl.: var. serie, res., aic (media, sqm)')
      write(6,40100) WVARt , WVARt2 
      write(6,40100) Wrest , Wrest2 
      write(6,40100) Waict , Waict2 
40100 format(1h    , 2F10.4)

      CLOSE(UNIT = 6)

      STOP

      END


      subroutine xsubst(p,mp,q,mq,indep,indeq,npzp,nqzq)
      implicit real*8 (a- h, o - z)
c        this routine yields the active lags of the true model
c        for the AR part in indep and for the MA part in indeq
      dimension  p(mp), q(mq), indep(mp), indeq(mq)
      data  toler/1.5D-5/
      ii = 0
	do 1  i =1 ,  mp
      if(dabs(p(i)) .gt. toler)  then
      ii =  ii + 1
      indep(ii)  =  I
      endif
1     continue
      npzp =  ii
      ii = 0
	do 2  i =1 ,  mq
      if(dabs(q(i)) .gt. toler)  then
      ii =  ii + 1
      indeq(ii)  =  I
      endif
2     continue
      nqzq =  ii
      return
      end
 

      SUBROUTINE ESTIMA(W, NHB, DM, RHO, COV, P, MP, Q, MQ,  R
     +, NAC, NAC1, TH, NX, NP, HDIFZ, HSIGNS, HQ, HP, HE, HPHI, HTB
     +, HDELZ, HA, HD, LSQFUN, EPS1, EPS2, MIT, NIT, NF, SSQ, IEXIT
     +, NPARAM, PARAM, AIC, IW, LIW, WI, LW) 

C     ESTIMATE OF SUBSET STATIONARY INVERTIBLE ARMA MODEL, N.COEF MP, MQ
C     INITIAL VALUES PARAM = (AR PARAM, MA PARAM), NPARAM = MP + MQ
C     THE AVERAGE PARAMETER AVEPA IS ALWAYS ESTIMATED, WHILST THE TREND
C     PARAMETER TREPA IS OPTIONAL: THE OUTPUT PARAMETER VECTOR IS TH(NP)
C     TH = (AVEPA, AR PARAMETERS, MA PARAMETERS, TREPA) 

      IMPLICIT DOUBLE PRECISION(A-H,O-Z)
      DIMENSION W(NHB), RHO(NAC1), COV(NAC1), P(MP), Q(MQ), R(NHB)
     +, PARAM(NPARAM)
      DIMENSION  HDIFZ(NX), HSIGNS(NX), HQ(NX), HP(NX), HE(NX), HPHI(NX)
     +, HTB(NX), HDELZ(LW), HA(LW), HD(LW), TH(NX)
     +, IW(LIW), WI(LW)                      
      COMMON/UNIIN/MFAC(3),AVEPA,ND(3),IOD(3),INC(6)                    
     *,            IOPA(10),UPA(10),ITREND ,TREPA                       
      COMMON/UNIOUT/NAR,NMA,NDIFAC,NARMAF,MAX1,MAX11,MAX2,MBO           
      COMMON/TSOBS/IWBF,NOB,X(5500)                                     
      DATA  TOLER/1.D-10/
      EXTERNAL LSQFUN

      IEXIT  = 0

      IP = 0
      DO  1 I =1  ,  MP
      COEF = PARAM(I)
      IF(DABS(COEF) .gt. TOLER) IP = IP + 1
1     P(I)  =  COEF
      IQ = 0
      I1  =  MP + 1
      I2 =  MP + MQ
      I = 0
      DO  2 J =  I1, I2
      COEF = PARAM(J)
      IF(DABS(COEF) .gt. TOLER) IQ = IQ + 1
      I = I + 1
2     Q(I)  =  COEF

c      WRITE(6, 201)
c201   FORMAT(1H   ,'... Estima: ')
c      WRITE(6, *)  'NPARAM = ', NPARAM
c      WRITE(6, 203) (PARAM(I), I =1 , NPARAM)
c      WRITE(6, *)  'MP,IP = ', MP, IP
c      WRITE(6, 203) (P(I), I =1 , MP)
c      WRITE(6, *)  'MQ,IQ = ', MQ, IQ
c      WRITE(6, 203) (Q(I), I =1 , MQ)

      IWBF = 1
      MFAC(1) = 1
      MFAC(2)  = 0
      MFAC(3)  =1
      NDIFAC=0
      KKMAX=MFAC(1)+MFAC(3)                                               
      IF(KKMAX .EQ. 0)GO TO 25
      IP  = 0
      DO  21  J  =1  ,  MP
      IF(DABS(P(J)) .GT. TOLER) THEN
      IP =  IP + 1
      IOPA(IP) =  J                                              
      UPA(IP) =  P(J)
      ENDIF
21    CONTINUE
      IQ =  0                                              
      DO  23  J  =MP+1, MP+MQ
      IF(DABS(Q(J-MP)) .GT. TOLER) THEN
      IQ = IQ  +1
      IOPA(IP+IQ) =  J - MP                                              
      UPA(IP+IQ) =  Q(J - MP)
      ENDIF
23    CONTINUE                                              
      INC(1)  =  IP
      INC(2)  =  IQ
      NARMAP=0                                                          
      DO 20 J=1,KKMAX                                                     
20    NARMAP=NARMAP+INC(J)
25    CONTINUE                                                   
      DO  26  I =1, NHB
26    X(I) =  W(I)

      CALL ZCOR(W,NHB,NAC,RHO,COV,DM,DSD,IEXIT)                     
      IF(IEXIT .NE. 0) THEN
      IEXIT = -99
      RETURN
      ENDIF

      AVEPA  =  DM
       
C      SETS UP PARAMETER VALUES OF LABELED COMMON /UNIOUT/              

       CALL INIT1(TH,NP,NDR,NHB,IWBF,IDF)                            

c       write(6 , 202)
c202    format(1h  , 'parameters initial guess')
c       write(6 ,  203)(th(j) , j =1  ,   np)
c203    format(7f10.4)
c       write(6 , 220)
c220    format(1h  , 'ARIMA model structure')
c      write(6,221)(MFAC(J),J=1,3)                                       
c221     format(1h ,'mfac = ', 3i10)
c        write(6 , 229)  ndifac
c229    format(1h  ,  'ndifac =  ',  i10)
c      IF(ndifac .NE. 0)write(6,222)(ND(J),IOD(J),J=1,KKMAX)               
c222     format(1h ,'nd, iod = ', 6i8)
c      IF(ndifac .eq. 0)write(6,223)  avepa                              
c223   format(1h  , 'avepa = ', D20.10)
c      write(6,224)(INC(J),J=1,6)                                        
c224     format(1h ,'inc = ', 6i8)
      kkmax  =  mfac(1) + mfac(3)
      NARMAP=0                                                          
      DO 320 J=1,kkMAX                                                  
320   NARMAP=NARMAP+INC(J)                                              
c      write(6,213)(IOPA(J),J=1,NARMAP)                                  
c213   format(1h   ,  10i7)
c      write(6,233)(UPA(J),J=1,NARMAP)                                   
c233   format(1h  ,  10f7.4)
c      write(6,225) ITREND                                               
c225   format(1h  , 'itrend  = ' , i10)
c      IF(ITREND .NE. 0)write(6, 226)TREPA                               
c226    format(1h  ,  'trepa  =  ', f14.4)
c      write(6, 227)  IWBF
c227    format(1h  'backforecasting: iwbf=1 yes, iwbf=0 no', i8)
c      write(6 , 228) MIT
c228   format(1h  ,'max iter Levenberg-Marquardt algo mit = ',i8)

C      LEVENBERG-MARQUARDT ALGORITHM FOR SUM OF SQUARES MINIMIZATION

      CALL GLM1(LSQFUN,IEXIT,NHB,NP,TH,R,SSQ                            
     *,               EPS1,EPS2,MIT,NIT,NF                              
     *,HDIFZ,HSIGNS,HQ,HP,HE,HPHI,HTB,HDELZ,HA,HD,IW,LIW,WI,LW)         

C      SETS UP PARAMETERS CURRENT ESTIMATES                             
C      TO INDIVIDUAL ARRAYS                                             

      CALL DEVPA1(TH,NP)
       AIC=DBLE(NHB)*DLOG(DABS(SSQ/DBLE(NHB)))+2D0*DBLE(NARMAP)
      RETURN

      END
                                                

      SUBROUTINE HYBRID(NPARAM,PARAM,N,T0,RHO,NT,BL,BU,POINTS,LOW,IFAULT
     +,NPARAM1,NX,NAC,NAC1,WHW,DM,RCORRE,COV,P,MP,Q,MQ,RES,TH,NHP
     +,HDIFZ,HSIGNS,HQ,HP,HE,HPHI,HTB,HDELZ,HA,HD,EPS1,EPS2,MIT,NIT,NF
     +,ITER2,SSQ,FUNCT,IEXIT,NHB,NMAX,PSBST,XCOEFF,XROOTS,AIC
     +,IW,LIW,W,LW)

C   Brooks,S.P. (1995) A hybrid optimization algorithm (Algorithm AS298)
C   {\it Applied Statistics},44, 530-552.
C
C   The last 3 rows of the calling sequence are the arguments of Estima
C   Both routines Estima and Hybrid minimize the objective function  

      IMPLICIT  REAL*8 (A - H, O- Z)
      IMPLICIT  INTEGER*4  (I -N)
      REAL*8  FUNCT
C     ..
C     .. Scalar Arguments ..
      DOUBLE PRECISION RHO,T0
      INTEGER*4 IFAULT,N,NP,NPARAM,NT
C     ..
C     .. Complex Arguments
      COMPLEX  XCOEFF(NPARAM1), XROOTS(NPARAM)
C     ..
C     .. Array Arguments ..
      DOUBLE PRECISION BL(NPARAM),BU(NPARAM),LOW(NPARAM1),
     +                 PARAM(NPARAM),POINTS(NMAX,N)
C     ..
C     .. Local Scalars ..
      DOUBLE PRECISION HBST,BST,DIFF,NEWOBJ,OBJF,ONE,P,T,U,ZERO
      INTEGER*4 I,I1,I2
C     ..
C     .. External Functions ..
c      DOUBLE PRECISION RANDOM
      REAL*8  RANDOM
      EXTERNAL RANDOM
C     ..
C     .. External Functions / Subroutines ..
c      EXTERNAL FUNCT1,TRADNL
        EXTERNAL  FUNCT , LSQFUN , ESTIMA
C     ..
C     .. Intrinsic Functions ..
      INTRINSIC DEXP
C     ..
C     .. Data and model arguments 
      DIMENSION WHW(NHB),RCORRE(NAC1),COV(NAC1),P(MP),Q(MQ),RES(NHB)
C     ..
C     .. Routine Estima arguments
      DIMENSION  HDIFZ(NX), HSIGNS(NX), HQ(NX), HP(NX), HE(NX), HPHI(NX)
     +, HTB(NX), HDELZ(LW), HA(LW), HD(LW), TH(NX)
     +, IW(LIW), W(LW)                      
C
C      Define the constants ZERO and ONE and set
C      initial values for BST and IFAULT
C
C     .. Data statements ..
      DATA ZERO,ONE,HBST/0.0D0,1.0D0,1.0D+10/
C     ..
      IFAULT = 0
      ITER   =  0
      ITER2  =  0
      BST = HBST
C
C      ******************************************
C      *** Start of Annealing component to    ***
C      *** generate a set of starting points. ***
C      ******************************************
C
C      Initialise the temperature
C
      T = T0
C
C      Choose an initial point
C
C      DO 10 I = 1,NPARAM
C          PARAM(I) = RANDOM(BL(I),BU(I))
C   10 CONTINUE
      CALL CHINIT(NPARAM,NPARAM1,PARAM,P,MP,Q,MQ,PSBST,BL,BU
     +,XCOEFF,XROOTS,IFAULT)
      IF(IFAULT  .NE. 0) RETURN
C
C      Calculate the function value for the initial point
C
C      CALL FUNCT1(NPARAM,PARAM,OBJF)
      OBJF= FUNCT(IEXIT,NHB,NPARAM,PARAM,P,MP,Q,MQ,RES,WHW,SSQ
     +, TH,NX,IW, LIW, W, LW)
C
C      Begin the loop over I1 until Nt Temp. reductions.
C
      DO 90 I1 = 1,NT
C
C      Set the counter for the No of points accepted to one
C
          NP = 1
C
C      Begin the loop over I2 until N points have been considered.
C
          DO 80 I2 = 1,N

          ITER  =  ITER  +  1
C
C      Record the present point
C
              DO 20 I = 1,NPARAM
                  POINTS(I,NP+1) = PARAM(I)
   20         CONTINUE
C
C      Select a new point
C
C              DO 30 I = 1,NPARAM
C                  PARAM(I) = RANDOM(BL(I),BU(I))
C   30         CONTINUE
      CALL CHINIT(NPARAM,NPARAM1,PARAM,P,MP,Q,MQ,PSBST,BL,BU
     +,XCOEFF,XROOTS,IFAULT)
      IF(IFAULT  .NE. 0) RETURN
C
C      Calculate the function value of this new point
C
C              CALL FUNCT1(NPARAM,PARAM,NEWOBJ)
      NEWOBJ= FUNCT(IEXIT,NHB,NPARAM,PARAM,P,MP,Q,MQ,RES,WHW,SSQ
     +, TH,NX,IW, LIW, W, LW)
C
C      Calculate the difference
C
              DIFF = (NEWOBJ-OBJF)
C
C      If the new point is better than the old one then increase the
C      Np counter, and record it as a possible starting point. Move
C      to this new point.
C
              IF (DIFF.LT.ZERO) THEN
                  NP = NP + 1
                  DO 40 I = 1,NPARAM
                      POINTS(I,NP) = PARAM(I)
   40             CONTINUE
                  OBJF = NEWOBJ
                  IF (NEWOBJ.LT.BST) THEN
                      BST = NEWOBJ

                         ITER2  =   ITER

                      DO 50 I = 1,NPARAM
                          POINTS(I,1) = PARAM(I)
   50                 CONTINUE
                  END IF
C
C      Otherwise use the Metropolis Criterion
C
              ELSE
C
C      P is the comparison statistic. U is a standard Uniform variate.
C
                  PHP = DEXP(-DIFF/T)
C                  U = RANDOM(ZERO,ONE)
         U = DMIN1(ONE,DMAX1(ZERO,ZERO+RANDOM(L)*(ONE-ZERO)))
C
C      If U < PHP then accept this point. Increase the Np counter,
C      Record this point as a possible starting point, and move
C      to this new point.
C
                  IF (U.LT.PHP) THEN
                      NP = NP + 1
                      OBJF = NEWOBJ
                      DO 60 I = 1,NPARAM
                          POINTS(I,NP) = PARAM(I)
   60                 CONTINUE
C
C      If U >= PHP then reject this point and return to the old point.
C
                  ELSE
                      DO 70 I = 1,NPARAM
                          PARAM(I) = POINTS(I,NP+1)
   70                 CONTINUE
                  END IF
C
C      Both alternatives for the comparison of the new and old point
C      Have been considered.
C
              END IF

C
C      Continue with the loop until N new points have been considered
C
   80     CONTINUE
C
C      Set the error indicator to 1 if the algorithm appears to have
C      converged.
C
          IF (NP.EQ.1) THEN
              IFAULT = I1
          END IF
C
C      Once N points have been considered, lower the Temperature.
C
          T = T*RHO
C
C      Continue with the loop until the Temp. has been reduced Nt times
C
   90 CONTINUE
C
C       *****************************************
C       *** End of Annealing component, NP    ***
C       *** starting points are now stored in ***
C       *** the array POINTS.                 ***
C       *****************************************
C
C       Set an initially high value of LOW(NPARAM+1)
C
      LOW(NPARAM+1) = HBST
C
C       *******************************************
C       *** Start of the traditional component, ***
C       *** which loops over the NP starting    ***
C       *** points calling the traditional      ***
C       *** minimisation routine for each.      ***
C       *******************************************
C
C       Begin the loop over the Np starting points.
C
      DO 130 I1 = 1,NP
C
C       Set the PARAM array to be the values of the I1st start point
C
          DO 110 I2 = 1,NPARAM
              PARAM(I2) = POINTS(I2,I1)
  110     CONTINUE
C
C       Call traditional minimisation routine
C
C          CALL TRADNL(NPARAM,PARAM,BL,BU,OBJF)
      CALL ESTIMA(WHW, NHB, DM, RCORRE, COV, P, MP, Q, MQ,  RES
     +, NAC, NAC1, TH, NX, NHP, HDIFZ, HSIGNS, HQ, HP, HE, HPHI, HTB
     +, HDELZ, HA, HD, LSQFUN, EPS1, EPS2, MIT, NIT, NF, SSQ, IEXIT
     +, NPARAM, PARAM, AIC,IW, LIW, W, LW) 
      J = 1     
      DO  930  I = 1 , NPARAM
      J = J  + 1
930   PARAM(I)  =  TH(J)    
c      OBJF= FUNCT(IEXIT,NHB,NPARAM,PARAM,P,MP,Q,MQ,RES,WHW,SSQL
c     +, TH,NX,IW, LIW, W, LW)
       objf = AIC
C
C       Check to see if this starting point gives the
C       lowest function value to date
C
          IF (OBJF.LT.LOW(NPARAM+1)) THEN
              DO 120 I = 1,NPARAM
                  LOW(I) = PARAM(I)
  120         CONTINUE
              LOW(NPARAM+1) = OBJF
              SSQL  =  SSQ
          END IF
C
C       End the Np loop
C
  130 CONTINUE
         SSQ  =   SSQL
C
C       **************************************
C       *** End of the second, traditional ***
C       *** component.                     ***
C       **************************************
C
C       Return to calling routine
C
      RETURN
C
C       End of subroutine HYBRID
C
      END


      SUBROUTINE CHINIT(NPARAM,NPARAM1,PARAM,P,MP,Q,MQ,PSBST,BL,BU
     +,XCOEFF,XROOTS,IFAULT)

C      CHOOSE AN INITIAL POINT, AR COEFFICIENTS P(1)..P(MP) AND MA
C      COEFFICIENTS Q(1)..Q(MQ): THE AR POLYNOMIAL HAS TO BE STATIONARY,
C      THE MA POLYNOMIAL HAS TO BE INVERTIBLE. SUBSET ARMA IS SIMULATED,
C      THAT IS A RANDOM PARAMETER IS GENERATED IF RANDOM(L) < PSBST,
C      OTHERWISE THE PARAMETER IS SET TO ZERO.
C      IFAULT =  -17 IF RANDOM GENERATION CANNOT FIT THE STATIONARITY
C      IFAULT = -317 IF RANDOM GENERATION CANNOT FIT THE INVERTIBILITY
C      IFAULT = -117 IF AR COEFFICIENTS ARE ALL ZERO  
C      IFAULT = -217 IF MA COEFFICIENTS ARE ALL ZERO  
C      IFAULT =    0 NO ERRORS

      IMPLICIT REAL*8 (A - H, O - Z)
      DIMENSION PARAM(NPARAM), P(MP), Q(MQ), BL(NPARAM), BU(NPARAM)
      COMPLEX  XCOEFF(NPARAM1), XROOTS(NPARAM)
      DATA  TOLER/1.5D-5/ , ZERO/0.0D-000/ , MAXIT /1000/
C      DATA  TOLER/10D-10/

      IFAULT = 0

C        GENERATE AR COEFFICIENTS

      ITER  = 0
1     CONTINUE

c       write(6, 107) iter
c107    format(1h   ,'... AR iter = ' , I5)

      IP = 0
      JP = 0
      DO  10  I  =1  , MP
      U  = RANDOM(L)
      IF(U .LT. PSBST)  THEN
      PARAM(I)= DMIN1(BU(I),DMAX1(BL(I),BL(I)+RANDOM(L)*(BU(I)-BL(I))))
      ELSE
      PARAM(I)  =  ZERO
      ENDIF
      IF(DABS(PARAM(I)) .GT. TOLER)  THEN
      IP  =  IP +1
      ENDIF
      JP = JP + 1
      P(JP) =  PARAM(I)
10    CONTINUE

c       write(6, 101)  psbst 
c101    format(1h  ,'generate AR parameters, psbst = ', F7.4)
c       write(6, 102) (p(i), i=1, mp)
c102    format(1h   ,7F10.6)

      IF(IP .LE. 0) THEN
      ITER = ITER + 1
      IF(ITER .GT. MAXIT) THEN
      IFAULT = -117
      RETURN
      ELSE
      GOTO  1
	ENDIF
      ENDIF

      CALL CHTRL(P,MP,ICONTR,XCOEFF,XROOTS)

c       write(6, 109) icontr
c109    format(1h   ,'... AR icontr = ' , I5)

      IF(ICONTR .LT. 0) THEN
      ITER = ITER + 1
      IF(ITER .GT. MAXIT) THEN
      IFAULT = -17
      RETURN
      ELSE
      GOTO  1
	ENDIF
      ENDIF

C        GENERATE MA COEFFICIENTS

      ITER  = 0
2     CONTINUE

c       write(6, 108) iter
c108    format(1h   ,'... MA iter = ' , I5)

      IQ = 0
      JQ = 0
      I1 = MP + 1
      DO  20  I  =I1  , NPARAM
      U  = RANDOM(L)
      IF(U .LT. PSBST)  THEN
      PARAM(I)= DMIN1(BU(I),DMAX1(BL(I),BL(I)+RANDOM(L)*(BU(I)-BL(I))))
      ELSE
      PARAM(I)  =  ZERO
      ENDIF
      IF(DABS(PARAM(I)) .GT. TOLER)  THEN
      IQ  =  IQ +1
      ENDIF
      JQ = JQ + 1
      Q(JQ) =  PARAM(I)
20    CONTINUE

c       write(6, 103)  psbst 
c103   format(1h  ,'generate MA parameters, psbst = ', F7.4)
c       write(6, 104) (q(i), i=1, mq)
c104   format(1h   ,7F10.6)
   
      IF(IQ .LE. 0) THEN
      ITER = ITER + 1
      IF(ITER .GT. MAXIT) THEN
      IFAULT = -217
      RETURN
      ELSE
      GOTO  2
	ENDIF
      ENDIF

      CALL CHTRL(Q,MQ,ICONTR,XCOEFF,XROOTS)

c       write(6, 110) icontr
c110    format(1h   ,'... MA icontr = ' , I5)

      IF(ICONTR .LT. 0) THEN
      ITER = ITER + 1
      IF(ITER .GT. MAXIT) THEN
      IFAULT = -317
      RETURN
      ELSE
      GOTO  2
	ENDIF
      ENDIF

c       write(6, 105) nparam
c105    format(1h  ,'vector param, nparam = ' , I5)
c       write(6, 106) (param(i), i=1, nparam)
c106   format(1h   ,7F10.6)

      RETURN
      END


      SUBROUTINE  CHTRL(PARVER, NQ, ICONTR, XCOEFF, XROOTS)
        implicit real*8 (a-h, o-z)
      COMPLEX  XCOEFF(NQ+1), XROOTS(NQ)
      LOGICAL polish
        dimension parver(NQ)
       lprint  =  6
      ONE =  1D0
        toler = 1.5D-5
c       check the actual order of the polynomial
         iq  =  0
        do  1001  i = 1 ,  nq
         if(dabs(parver(nq - i + 1)) .gt. toler) then
         iq = nq - i + 1
         goto  1002
         endif
1001    continue
1002    continue 
c        check the roots of polynomial 1-p1B-p2B^2... 
c        all outside the unit circle
        icontr  =  -1
       if(iq  .le.  0)  return
      XCOEFF(1) =CMPLX(one)
      DO  1400  J  =  1  ,  iQ
      RCOEFF =  -parver(J)
      JJ  =  J + 1
      XCOEFF(JJ) =CMPLX(RCOEFF)
1400  continue
c      write(lprint, *) (xcoeff(j), j=1 , iq+1)
      polish=.false.
      call zroots(XCOEFF,iQ,XROOTS,polish)
c      write(lprint, *) (xroots(j), j=1 , iq)
      do  1015 i  =  1  ,  iQ
      xnorm = CABS(xroots(i))
c      write(lprint,'(1x,i4,4x, F10.4)') i, xnorm
      if(xnorm .le. one) goto  1409  
1015  continue
       icontr  =  0
1409   continue
      return
      end


      REAL*8 FUNCTION FUNCT(IEXIT,N,NPARAM,PARAM,P,MP,Q,MQ,E,W,SUMSQ
     +, TH, NX, IW, LIW, WI, LW)

C     ARMA(MP,MQ) RESIDUAL EVALUATION: QUICK RECURSION + AIC CRITERION

      IMPLICIT REAL*8 (A - H, O - Z)
      DIMENSION PARAM(NPARAM), W(N), E(N), P(MP), Q(MQ), TH(NX)
     *,            IW(LIW), WI(LW)
      COMMON/UNIIN/MFAC(3),AVEPA,ND(3),IOD(3),INC(6)                    
     *,            IOPA(10),UPA(10),ITREND ,TREPA                       
      COMMON/UNIOUT/NAR,NMA,NDIFAC,NARMAF,MAX1,MAX11,MAX2,MBO           
      COMMON/TSOBS/IWBF,NOB,X(5500)                                     
C      DATA  TOLER/10D-10/
      DATA  TOLER/1.5D-5/

C      DO  4466 I = 1, NOB
C4466  X(I) =  W(I)
C      TH(1) =  AVEPA
C      DO  4455 I = 1, NPARAM
C4455  TH(I + 1) = PARAM(I)
C      NP = NPARAM+1
C      SUMSQ = 0                                                         ARI10860
C      CALL LSQFUN(IEXIT,NOB,NP,TH,E,IW,LIW,WI,LW)                       ARI10870
C      DO 90 I = 1, NOB                                                  ARI10890
C   90 SUMSQ=SUMSQ+E(I)*E(I)                                             ARI10900
C      IEXIT  = 0
C       WRITE(*,*) NOB,SUMSQ,PARAM(5)
      IP = 0
      DO  1 I =1  ,  MP
      COEF = PARAM(I)
      IF(DABS(COEF) .gt. TOLER) IP = IP + 1
1     P(I)  =  COEF
      IQ = 0
      I1  =  MP + 1
      I2 =  MP + MQ
      I = 0
      DO  2 J =  I1, I2
      COEF = PARAM(J)
      IF(DABS(COEF) .gt. TOLER) IQ = IQ + 1
      I = I + 1
2     Q(I)  =  COEF

c         WRITE(6, *)  '...'
c         WRITE(6, *)  'ENTER  FUNCT'
c         WRITE(6, *)  'NPARAM = ', NPARAM
c         WRITE(6, *) (PARAM(I), I =1 , NPARAM)
c         WRITE(6, *)  'MP,IP = ', MP, IP
c         WRITE(6, *) (P(I), I =1 , MP)
c         WRITE(6, *)  'MQ,IQ = ', MQ, IQ
c         WRITE(6, *) (Q(I), I =1 , MQ)
        
            SUMSQ = 0.0D-000
            MXPQ = MAX0(MP, MQ)
            DO 380 I = MXPQ, N
              E(I) = W(I)
              IF (MP .LE. 0) GO TO 340
              DO 321 J = 1, MP
                IMJ = I - J
                E(I) = E(I) - P(J) * W(IMJ)
  321        CONTINUE
  340        IF (MQ .LE. 0) GO TO 380
              DO 350 J = 1, MQ
                IMJ = I - J
                E(I) = E(I) + Q(J) * E(IMJ)
  350        CONTINUE
  380      SUMSQ = SUMSQ + E(I) * E(I)

      FUNCT = DBLE(N) * DLOG(DABS(SUMSQ)/DBLE(N)) + 2D0 * DBLE(IP + IQ)

c           WRITE(6, *) 'SUMSQ = ', SUMSQ
c           WRITE(6, *) 'N = ', N
c           WRITE(6, *) 'RES.VAR = ', SUMSQ/DBLE(N) 
c           WRITE(6, *) 'AIC (FUNCT) = ', FUNCT
c         WRITE(6, *)  'EXIT  FUNCT'
c          WRITE(6, *)  '...'

         RETURN
          END


      SUBROUTINE SIMTRU(W,N,NSCART,NTOT,Z,ETILDA,P,NP,Q,NQ,VARE)

C       ARMA MODEL SIMULATION

      IMPLICIT REAL*8 (A -H , O - Z)
      DIMENSION W(N),Z(NTOT),ETILDA(NTOT),P(NP),Q(NQ) 

C       WHITE NOISE GENERATION N(0,VARE)

      DO  4401  I  =  1  ,  NTOT
4402   PPX0 =  RANDOM(L)
      PPX1 =  PPND16(PPX0 , IZAULT)
      IF(IZAULT .NE. 0)  GOTO  4402
4401   ETILDA(I) =  PPX1 * DSQRT(VARE)   

C       WHITE NOISE FROM 1 TO NM=MAX(NP,NQ)

        NM = MAX0(NP,NQ)
        DO  2 I =1 , NM
2       Z(I)   =  ETILDA(I)

C       ARMA MODEL FROM NM+1 TO NTOT

      NM1 =  NM+ 1
      DO  3  I = NM1 , NTOT
      Z(I) = ETILDA(I) 
      IF(NP .GT. 0)  THEN
      DO  6   J =  1  , NP
      Z(I) = Z(I) + P(J) * Z(I - J)
6     CONTINUE
      ENDIF
      IF(NQ .GT. 0)  THEN
      DO  8   J =  1  , NQ
      Z(I) = Z(I) - Q(J) * ETILDA(I - J)
8     CONTINUE
      ENDIF
3     CONTINUE

C       END ARMA MODEL SIMULATION	       

      II =  NSCART
      DO  1  I  =1  ,  N
      II  =  II  + 1
1     W(I)  =  Z(II)

      RETURN
      END


      REAL*8 function  random(l)

c        algorithm as 183 Appl. Statist. (1982) Vol. 31, no. 2
c
c        returns a pseudo-random number rectangularly distributed 
c        between  0  and 1.
c
c        ix, iy and iz should be set to integer values between 
c        1 and 30000 before first entry
c
c        integer arithmetic up to 30323  is required

      IMPLICIT REAL*8 (A -H , O - Z)
	common /rand/  ix, iy,  iz
	ix = 171 * mod(ix, 177) - 2 * (ix / 177)
	iy = 172 * mod(iy, 176) - 35 * (iy / 176)
	iz = 170 * mod(iz, 178) - 63 * (iz / 178)

	if (ix .lt. 0) ix = ix + 30269
	if (iy .lt. 0) iy = iy + 30307
	if (iz .lt. 0) iz = iz + 30323

c        if integer arithmetic up to 5212632 is available, 
c        the preceding 6 statements may be replaced by
c
c        ix = mod(171 * ix, 30269)
c        iy = mod(172 * iy, 30307) 
c        iz = mod(170 * iz, 30323)
c
c        on some machines, this may slightly increase
c        the speed. the results will be identical.

	random  = DMOD(DBLE(ix) / 30269D0 + DBLE(iy) / 30307D0 +
     *                 DBLE(iz) / 30323D0, 1D0)

	return
	end


        DOUBLE PRECISION FUNCTION PPND16 (P, IFAULT)
C
C       ALGORITHM AS241  APPL. STATIST. (1988) VOL. 37, NO. 3
C
C       Produces the normal deviate Z corresponding to a given lower
C       tail area of P; Z is accurate to about 1 part in 10**16.
C
C       The hash sums below are the sums of the mantissas of the
C       coefficients.   They are included for use in checking
C       transcription.
C
        DOUBLE PRECISION ZERO, ONE, HALF, SPLIT1, SPLIT2, CONST1,
     *          CONST2, A0, A1, A2, A3, A4, A5, A6, A7, B1, B2, B3,
     *          B4, B5, B6, B7,
     *          C0, C1, C2, C3, C4, C5, C6, C7, D1, D2, D3, D4, D5,
     *          D6, D7, E0, E1, E2, E3, E4, E5, E6, E7, F1, F2, F3,
     *          F4, F5, F6, F7, P, Q, R
        PARAMETER (ZERO = 0.D0, ONE = 1.D0, HALF = 0.5D0,
     *          SPLIT1 = 0.425D0, SPLIT2 = 5.D0,
     *          CONST1 = 0.180625D0, CONST2 = 1.6D0)
C
C       Coefficients for P close to 0.5
C
        PARAMETER (A0 = 3.38713 28727 96366 6080D0,
     *             A1 = 1.33141 66789 17843 7745D+2,
     *             A2 = 1.97159 09503 06551 4427D+3,
     *             A3 = 1.37316 93765 50946 1125D+4,
     *             A4 = 4.59219 53931 54987 1457D+4,
     *             A5 = 6.72657 70927 00870 0853D+4,
     *             A6 = 3.34305 75583 58812 8105D+4,
     *             A7 = 2.50908 09287 30122 6727D+3,
     *             B1 = 4.23133 30701 60091 1252D+1,
     *             B2 = 6.87187 00749 20579 0830D+2,
     *             B3 = 5.39419 60214 24751 1077D+3,
     *             B4 = 2.12137 94301 58659 5867D+4,
     *             B5 = 3.93078 95800 09271 0610D+4,
     *             B6 = 2.87290 85735 72194 2674D+4,
     *             B7 = 5.22649 52788 52854 5610D+3)
C       HASH SUM AB    55.88319 28806 14901 4439
C
C       Coefficients for P not close to 0, 0.5 or 1.
C
        PARAMETER (C0 = 1.42343 71107 49683 57734D0,
     *             C1 = 4.63033 78461 56545 29590D0,
     *             C2 = 5.76949 72214 60691 40550D0,
     *             C3 = 3.64784 83247 63204 60504D0,
     *             C4 = 1.27045 82524 52368 38258D0,
     *             C5 = 2.41780 72517 74506 11770D-1,
     *             C6 = 2.27238 44989 26918 45833D-2,
     *             C7 = 7.74545 01427 83414 07640D-4,
     *             D1 = 2.05319 16266 37758 82187D0,
     *             D2 = 1.67638 48301 83803 84940D0,
     *             D3 = 6.89767 33498 51000 04550D-1,
     *             D4 = 1.48103 97642 74800 74590D-1,
     *             D5 = 1.51986 66563 61645 71966D-2,
     *             D6 = 5.47593 80849 95344 94600D-4,
     *             D7 = 1.05075 00716 44416 84324D-9)
C       HASH SUM CD    49.33206 50330 16102 89036
C
C       Coefficients for P near 0 or 1.
C
        PARAMETER (E0 = 6.65790 46435 01103 77720D0,
     *             E1 = 5.46378 49111 64114 36990D0,
     *             E2 = 1.78482 65399 17291 33580D0,
     *             E3 = 2.96560 57182 85048 91230D-1,
     *             E4 = 2.65321 89526 57612 30930D-2,
     *             E5 = 1.24266 09473 88078 43860D-3,
     *             E6 = 2.71155 55687 43487 57815D-5,
     *             E7 = 2.01033 43992 92288 13265D-7,
     *             F1 = 5.99832 20655 58879 37690D-1,
     *             F2 = 1.36929 88092 27358 05310D-1,
     *             F3 = 1.48753 61290 85061 48525D-2,
     *             F4 = 7.86869 13114 56132 59100D-4,
     *             F5 = 1.84631 83175 10054 68180D-5,
     *             F6 = 1.42151 17583 16445 88870D-7,
     *             F7 = 2.04426 31033 89939 78564D-15)
C       HASH SUM EF    47.52583 31754 92896 71629
C
        IFAULT = 0
        Q = P - HALF
        IF (ABS(Q) .LE. SPLIT1) THEN
          R = CONST1 - Q * Q
          PPND16 = Q * (((((((A7 * R + A6) * R + A5) * R + A4) * R + A3)
     *                  * R + A2) * R + A1) * R + A0) /
     *                (((((((B7 * R + B6) * R + B5) * R + B4) * R + B3)
     *                  * R + B2) * R + B1) * R + ONE)
          RETURN
        ELSE
          IF (Q .LT. ZERO) THEN
            R = P
          ELSE
            R = ONE - P
          END IF
          IF (R .LE. ZERO) THEN
            IFAULT = 1
            PPND16 = ZERO
            RETURN
          END IF
          R = SQRT(-LOG(R))
          IF (R .LE. SPLIT2) THEN
            R = R - CONST2
            PPND16 = (((((((C7 * R + C6) * R + C5) * R + C4) * R + C3)
     *                  * R + C2) * R + C1) * R + C0) /
     *               (((((((D7 * R + D6) * R + D5) * R + D4) * R + D3)
     *                  * R + D2) * R + D1) * R + ONE)
          ELSE
            R = R - SPLIT2
            PPND16 = (((((((E7 * R + E6) * R + E5) * R + E4) * R + E3)
     *                  * R + E2) * R + E1) * R + E0) /
     *               (((((((F7 * R + F6) * R + F5) * R + F4) * R + F3)
     *                  * R + F2) * R + F1) * R + ONE)
          END IF
          IF (Q .LT. ZERO) PPND16 = - PPND16
          RETURN
        END IF
        END


      SUBROUTINE GLM1(LSQFUN,IEXIT,NOB,NP,TH,R,SSQ                      ARI09510
     *,               EPS1,EPS2,MIT,NIT,NF                              ARI09520
     *,DIFZ,SIGNS,Q,P,E,PHI,TB,DELZ,A,D,IW,LIW,W,LW)                    ARI09530
C                                                                       ARI09540
C ---- SUBROUTINE GLM1                                                  ARI09550
C                                                                       ARI09560
C     ROUTINE OF UNCONSTRAINED OPTIMIZATION                             ARI09570
C     USING GAUSS - LEVENBERG - MARQUARDT METHOD                        ARI09580
C     AS IMPLEMENTED BY D.W.MARQUARDT (1963)                            ARI09590
C     "AN ALGORITHM FOR LEAST -SQUARES ESTIMATION                       ARI09600
C     OF NON LINEAR PARAMETERS",SIAM J.VOL.11,                          ARI09610
C     P.431-441.                                                        ARI09620
C                                                                       ARI09630
C ----                                                                  ARI09640
C                                                                       ARI09650
C     THIS PROGRAM IS AN ADAPTATION OF THE NON LINEAR                   ARI09660
C     LEAST SQUARES ALGORITHM AS DEVELOPPED IN THE UNSUPPORTED MACC     ARI09670
C     PROGRAM 'GAUSHAUS', ORIGINALLY WRITTEN BY D.A. MEETER(1964)       ARI09680
C     'PROGRAM GAUSHAUS',NUMERICAL ANALYSIS LABORATORY,                 ARI09690
C     UNIVERSITY OF WISCONSIN, MADISON ( REVISED , 1966 )               ARI09700
C                                                                       ARI09710
C     ALSO REFERENCED IN G.E.P. BOX - G.M.JENKINS(1976) TIME SERIES     ARI09720
C     ANALYSIS  FORECASTING AND CONTROL , HOLDEN DAY,PART V P.504.      ARI09730
C                                                                       ARI09740
C     ALSO REFERENCED IN AFS PACKAGE BY D.J.PACK (1979)                 ARI09750
C     'A COMPUTER PROGRAM FOR THE ANALYSIS OF TIME SERIES MODELS        ARI09760
C     USING THE BOX-JENKINS PHILOSOPHY' , AUTOMATIC FORECASTING         ARI09770
C     SYSTEMS, INC., PO BOX 563, HATBORO,PA 19040.                      ARI09780
C                                                                       ARI09790
C     ALSO REFERENCED IN WMTS-1 PACKAGE BY M.R. GROUPE,G.B. HUDAK, W.R. ARI09800
C     G.C. TIAO, IH CHANG, G.E.P. BOX (1979)WISCONSIN MULTIPLE TIME     ARI09810
C     SERIES PROGRAM PACKAGE ,DEPARTMENT OF STATISTICS,UNIVERSITY OF    ARI09820
C     WISCONSIN , MADISON.                                              ARI09830
C                                                                       ARI09840
C     ALSO REFERENCED IN BMDP PACKAGE ,UCLA ,P-SERIES ,1981 LOS ANGELES ARI09850
C     PROGRAM P2T BOX-JENKINS TIME SERIES ANALYSIS.                     ARI09860
C                                                                       ARI09870
C ----                                                                  ARI09880
C                                                                       ARI09890
C     TH  REAL ARRAY OF DIMENSION AT LEAST NP.BEFORE ENTRY, TH(J) MUST  ARI09900
C         BE SET TO A GUESS OF THE J(TH) COMPONENT OF THE POSITION OF THARI09910
C         MINIMUM (J=1,2,....,NP) ON EXIT, TH CONTAINS THE  POINT FOUND ARI09920
C         DURING THE CALCULATIONS.                                      ARI09930
C                                                                       ARI09940
C     NP  ON ENTRY ,NP SPECIFIES  THE NUMBER OF PARAMETERS. UNCHANGED ONARI09950
C                                                                       ARI09960
C     R  REAL ARRAY OF DIMENSION AT LEAST NDR.ON EXIT , R CONTAINS MODELARI09970
C         RESIDUALS AT THE POINT WHICH YIELDS THE FINAL VALUES OF  SSQ. ARI09980
C                                                                       ARI09990
C     NOB  ON ENTRY , NOB SPECIFIES THE NUMBER OF RESIDUALS, UNCHANGED OARI10000
C         EXIT                                                          ARI10010
C                                                                       ARI10020
C     W   REAL ARRAY OF DIM LW AT LEAST MAX( 2*(2*NOB+1)+5*200,NP*NP)   ARI10030
C         WORKSPACE TO BE USED AS REFERENCED BY  ROUTINE BCKF           ARI10040
C         ON EXIT, THE NP*NP FIRST POSITIONS OF W CONTAIN THE CORRELATIOARI10050
C          MATRIX OF THE PARAMETERS                                     ARI10060
C                                                                       ARI10070
C     MIT ON ENTRY, MIT SPECIFIES THE MAXIMUM NUMBER OF ITERATION       ARI10080
C        THAT ARE  ALLOWED.UNCHANGED ON EXIT.                           ARI10090
C                                                                       ARI10100
C     EPS1 IF RELATIVE CHANGE IN RESIDUAL SUM OF SQUARES LESS THAN THIS ARI10110
C          THE ITERATIVE  CHANGE IN PARAMETER VALUES STOPS AT PRESUMED OARI10120
C                                                                       ARI10130
C     EPS2  IF RELATIVE CHANGE IN EACH INDIVIDUAL PARAMETER LESS THAN THARI10140
C          THE ITERATIVE CHANGE IN PARAMETER VALUES  STOPS AT PRESUMED OARI10150
C                                                                       ARI10160
C     SSQ ON EXIT ,SSQ CONTAINS THE VALUE OF THE SUM OF SQUARES CORRESPOARI10170
C         THE FINAL POINT STORED IN TH.                                 ARI10180
C                                                                       ARI10190
C      NIT GIVES THE NUMBER OF  ITERATIONS WHICH HAVE BEEN PERFORMED IN ARI10200
C                                                                       ARI10210
C     NF SPECIFIES THE NUMBER OF TIMES THAT RESIDUAL EVALUATION  ROUTINEARI10220
C         HAS BEEN CALLED TO FAR.                                       ARI10230
C                                                                       ARI10240
C     DIFZ,SIGNS,Q,P,E,PHI,TB REAL ARRAY OF DIMENSION AT LEAST (NP) USEDARI10250
C     WORKSPACE.                                                        ARI10260
C                                                                       ARI10270
C     DELZ  REAL ARRAY OF DIMENSION AT LEAST NOB*NP. USED AS WORKSPACE. ARI10280
C     (CURRENT JACOBIAN MATRIX IS STORED IN DELZ)                       ARI10290
C                                                                       ARI10300
C     A,D REAL ARRAYS OF DIMENSION AT LEAST NP*NP. USED AS WORKSPACES   ARI10310
C                                                                       ARI10320
C     IW  INTEGER ARRAY OF DIMENSION (LIW). NOT USED IN THIS ROUTINE    ARI10330
C                                                                       ARI10340
C     IEXIT: UNLESS THE ROUTINE DETECTS AN ERROR  IEXIT CONTAINS 1 ON EXARI10350
C                                                                       ARI10360
C ----                                                                  ARI10370
C                                                                       ARI10380
      IMPLICIT REAL*8(A-H , O-Z)
      DIMENSION TH(NP),R(NOB),IW(LIW),W(LW)                             ARI10390
     +,DIFZ(NP),SIGNS(NP),Q(NP),P(NP),E(NP),PHI(NP),TB(NP)       
     *,DELZ(LW),A(LW),D(LW)                                             ARI10400
C                                                                       ARI10410
C SET INITIAL VALUES OF LAMBDA AND  FNU AS DEFINED IN MARQUARDT'S ALG.  ARI10420
C                                                                       ARI10430
      DATA FLAM /0.1/,FNU/10.0/                                         ARI10440
      DATA NPRINT/6/                                                    ARI10450
C                                                                       ARI10460
C     MAXCAL = MAX NUMBER OF TIME THE STEP IS HALVED                    ARI10470
C     MAXCNT = MAX NUMBER OF TIME THE STEP IS RESTRICTED                ARI10480
C     MAXBF  = MAX NUMBER OF ITERATIONS OF BACKFORECASTING              ARI10490
C                                                                       ARI10500
             DATA MAXCAL/5/,MAXCNT/36/,MAXBF/5/                         ARI10510
C                                                                       ARI10520
C     SET INITIAL VALUES  OF EPSILON(J) (DIFZ(J)) TO BE USED IN FINITE  ARI10530
C     DERIVATIVES APPROXIMATION. SET INITIAL VALUES OF SIGNS ARRAY TO   ARI10540
C     APPROXIMATED ZERO.                                                ARI10550
C                                                                       ARI10560
      JJ=0                                                              ARI10570
      JI=1                                                              ARI10580
      DO 400 J=1,NP                                                     ARI10590
      DIFZ(J)=0.01+FLOAT(JJ)*(-.01)                                     ARI10600
400   SIGNS(J)=0.01+FLOAT(JI)*(-.01)                                    ARI10610
C                                                                       ARI10620
C     SET INITIAL VALUES OF NIT, NF , IEXIT  AND  GA                    ARI10630
C                                                                       ARI10640
      NIT=0                                                             ARI10650
      NF=0                                                              ARI10660
      IEXIT=0                                                           ARI10670
      GA = FLAM                                                         ARI10680
C----     ................................................              ARI10690
      ASSIGN 225 TO IRAN                                                ARI10700
      ASSIGN 265 TO JORDAN                                              ARI10710
      IF(EPS1) 5, 10, 10                                                ARI10720
 5    EPS1 = 0                                                          ARI10730
 10   IF(EPS2) 40, 40, 30                                               ARI10740
 40   IF(EPS1) 60, 60, 50                                               ARI10750
   60 ASSIGN 270 TO IRAN                                                ARI10760
      GO TO 70                                                          ARI10770
   50 ASSIGN 265 TO IRAN                                                ARI10780
      GO TO 70                                                          ARI10790
 30   IF(EPS1) 80, 80, 70                                               ARI10800
   80 ASSIGN 270 TO JORDAN                                              ARI10810
C----     ...............................................               ARI10820
C                                                                       ARI10830
C     INITIAL SUM OF SQUARE CALCULATION                                 ARI10840
C                                                                       ARI10850
   70 SSQ = 0                                                           ARI10860
      CALL LSQFUN(IEXIT,NOB,NP,TH,R,IW,LIW,W,LW)                        ARI10870
      NF=NF+1                                                           ARI10880
      DO 90 I = 1, NOB                                                  ARI10890
   90 SSQ=SSQ+R(I)*R(I)                                                 ARI10900
c     WRITE(NPRINT, 1003)SSQ                                            ARI10910
c1003 FORMAT(5X,'ISOS',E11.4)                                           ARI10920
      IF(NP .EQ. 0)GO TO 284                                            ARI10930
C                                                BEGIN ITERATION        ARI10940
100   NIT=NIT+1                                                         ARI10950
      GA = GA / FNU                                                     ARI10960
      INTCNT = 0                                                        ARI10970
 101  JS = 1 - NOB                                                      ARI10980
      DO 130 J=1,NP                                                     ARI10990
      JS= JS + NOB                                                      ARI11000
      TEMP = TH(J)                                                      ARI11010
      P(J)=DIFZ(J)*TH(J)                                                ARI11020
      IFAIL=0                                                           ARI11030
 105  TH(J)=TH(J)+P(J)                                                  ARI11040
      Q(J)=0                                                            ARI11050
C                                                                       ARI11060
C             EVALUATE GRADIENT AT TH(J),J=1,NP                         ARI11070
C                                                                       ARI11080
      CALL LSQFUN(IEXIT,NOB,NP,TH,DELZ(JS),IW,LIW,W,LW)                 ARI11090
      NF=NF+1                                                           ARI11100
      IF(IEXIT .EQ. 0)GO TO 110                                         ARI11110
C     AN ATTEMPT IS MADE TO FORCE CONVERGENCE OF BACKFORECASTING        ARI11120
C     PROCESS BY HALVING THE STEP (UP TO 5 TIMES)                       ARI11130
      IFAIL=IFAIL+1                                                     ARI11140
      IF(IFAIL.GT.MAXBF)GOTO 9004                                       ARI11150
      P(J)=P(J)/2.0                                                     ARI11160
      TH(J)=TEMP                                                        ARI11170
      GO TO 105                                                         ARI11180
 110  IJ=JS-1                                                           ARI11190
      DO 120 I = 1, NOB                                                 ARI11200
      IJ = IJ + 1                                                       ARI11210
      DELZ(IJ)=     -DELZ(IJ)+R(I)                                      ARI11220
 120  Q(J) = Q(J) + DELZ(IJ) * R(I)                                     ARI11230
      Q(J)= Q(J)/P(J)                                                   ARI11240
C                                                                       ARI11250
C                                      Q=XT*R (STEEPEST DESCENT)        ARI11260
C                                                                       ARI11270
  130 TH(J) = TEMP                                                      ARI11280
      DO 150 I = 1, NP                                                  ARI11290
      DO 151 J=1,I                                                      ARI11300
      SUM = 0                                                           ARI11310
      KJ = NOB*(J-1)                                                    ARI11320
      KI = NOB*(I-1)                                                    ARI11330
      DO 160 K = 1, NOB                                                 ARI11340
      KI = KI + 1                                                       ARI11350
      KJ = KJ + 1                                                       ARI11360
 160  SUM = SUM + DELZ(KI) * DELZ(KJ)                                   ARI11370
      TEMP= SUM/(P(I)*P(J))                                             ARI11380
      JI = J + NP*(I-1)                                                 ARI11390
      D(JI) = TEMP                                                      ARI11400
      IJ = I + NP*(J-1)                                                 ARI11410
 151  D(IJ) = TEMP                                                      ARI11420
 150  E(I) = DSQRT(D(JI))                                               ARI11430
 666  CONTINUE                                                          ARI11440
      DO 153  I = 1, NP                                                 ARI11450
      IJ = I-NP                                                         ARI11460
      DO 153 J=1,I                                                      ARI11470
      IJ = IJ + NP                                                      ARI11480
      A(IJ) = D(IJ) / (E(I)*E(J))                                       ARI11490
      JI = J + NP*(I-1)                                                 ARI11500
 153  A(JI) = A(IJ)                                                     ARI11510
C                                                                       ARI11520
C                                      A= SCALED MOMENT MATRIX          ARI11530
C                                                                       ARI11540
      II = - NP                                                         ARI11550
      DO 155 I=1,NP                                                     ARI11560
      P(I)=Q(I)/E(I)                                                    ARI11570
      PHI(I)=P(I)                                                       ARI11580
      II = NP + 1 + II                                                  ARI11590
 155  A(II) = A(II) + GA                                                ARI11600
C                                                                       ARI11610
      I=1                                                               ARI11620
       CALL MATIN(A, NP, P, I, DET)                                     ARI11630
C                                                                       ARI11640
C                                      P/E = CORRECTION VECTOR          ARI11650
C                                                                       ARI11660
       STEP=1.0                                                         ARI11670
      SUM1=0.                                                           ARI11680
      SUM2=0.                                                           ARI11690
      SUM3=0.                                                           ARI11700
      DO 231 I=1,NP                                                     ARI11710
      SUM1=P(I)*PHI(I)+SUM1                                             ARI11720
      SUM2=P(I)*P(I)+SUM2                                               ARI11730
      SUM3= PHI(I) * PHI(I) + SUM3                                      ARI11740
 231  PHI(I) = P(I)                                                     ARI11750
       TEMP = SUM1/DSQRT(SUM2*SUM3)                                     ARI11760
      TEMP = DMIN1(TEMP, 1D0)                                           ARI11770
       TEMP=57.295*DCOS(TEMP)                                           ARI11780
      IFAIL=0                                                           ARI11790
      IFAIL1=0                                                          ARI11800
 170  DO 220 I = 1, NP                                                  ARI11810
      P(I) = PHI(I) *STEP / E(I)                                        ARI11820
      TB(I) = TH(I) + P(I)                                              ARI11830
 220  CONTINUE                                                          ARI11840
c      WRITE(NPRINT,7000)(TB(I),I=1,NP)                                 ARI11850
c7000 FORMAT(5X,'TPPV',5E12.4/9X,5E12.4)                                ARI11860
C                                                                       ARI11870
C     TEST DIRECTIONS TB AND TH                                         ARI11880
C     Y=SIGN(X1,X2) IS THE TRANSFER OF SIGN FUNCTION                    ARI11890
C     Y=ABS(X1) IF X2>=0                                                ARI11900
C     Y=-ABS(X1) IF X2<0                                                ARI11910
C                                                                       ARI11920
C     IF SOME OF THE COMPONENTS OF TB AND TH HAVE OPPOSITE SIGNS,       ARI11930
C     THEN TEST 'TEMP' ANGLE  .    IF  TEMP  IS GREATER THAN THE        ARI11940
C     CRITERION ANGLE OF 30 DEGREES, THEN SET GA=4*GA AND GOTO 666 .    ARI11950
C     IF TEMP IS LESS THAN 30 DEGREES, THEN THE STEP IS HALVED.         ARI11960
C     (SEE D.W.MARQUARDT(1963): MARQUARDT USES A CRITERION ANGLE        ARI11970
C      OF 45 DEGREES)                                                   ARI11980
C                                                                       ARI11990
      DO 221 I = 1, NP                                                  ARI12000
      IF(SIGNS(I)) 221, 221, 222                                        ARI12010
 222  IF(DSIGN(1D0,TH(I))*DSIGN(1D0,TB(I))) 663, 221, 221               ARI12020
 221  CONTINUE                                                          ARI12030
      SUMB=0                                                            ARI12040
      CALL LSQFUN(IEXIT,NOB,NP,TB,R,IW,LIW,W,LW)                        ARI12050
      NF=NF+1                                                           ARI12060
      IF(IEXIT .EQ. 0)GO TO 223                                         ARI12070
C                                                                       ARI12080
C     AN ATTEMPT IS MADE TO FORCE CONVERGENCE OF BACKFORECASTING        ARI12090
C     PROCESS BY HALVING THE STEP (UP TO 5 TIMES)                       ARI12100
C                                                                       ARI12110
      IFAIL=IFAIL+1                                                     ARI12120
      IF(IFAIL.GT.MAXBF)GOTO 9004                                       ARI12130
      STEP=STEP/2.0                                                     ARI12140
      GO TO 170                                                         ARI12150
 223  DO 230 I=1,NOB                                                    ARI12160
230   SUMB=SUMB+R(I)*R(I)                                               ARI12170
      IF(SUMB - (1.0+EPS1)*SSQ) 662, 662, 663                           ARI12180
 663  IF( DMIN1(TEMP-30.0, GA)) 665, 665, 664                           ARI12190
665   STEP=STEP/2.0                                                     ARI12200
      IFAIL1 = IFAIL1 + 1                                               ARI12210
      IF(IFAIL1 - MAXCAL) 170, 9005, 9005                               ARI12220
  664 GA=GA*FNU                                                         ARI12230
      INTCNT = INTCNT + 1                                               ARI12240
      IF(INTCNT - MAXCNT) 666, 9005, 9005                               ARI12250
  662 CONTINUE                                                          ARI12260
      DO 669 I=1,NP                                                     ARI12270
669   TH(I)=TB(I)                                                       ARI12280
c      WRITE(NPRINT,2007)(TH(I),I=1,NP)                                 ARI12290
c2007 FORMAT(5X,'PVVR',5E12.4/9X,5E12.4)                                ARI12300
c      WRITE(NPRINT,1040)SUMB                                           ARI12310
c1040 FORMAT(5X,'SSAR',E14.7)                                           ARI12320
      GO TO IRAN, (225,265,270)                                         ARI12330
 225  DO 240 I = 1, NP                                                  ARI12340
      IF(DABS(P(I))/(1.D-20+DABS(TH(I)))-EPS2) 240, 240, 241            ARI12350
 241  GO TO JORDAN, (265,270)                                           ARI12360
  240 CONTINUE                                                          ARI12370
      IEXIT=1                                                           ARI12380
      GO TO 280                                                         ARI12390
 265  IF(DABS(SUMB - SSQ) - EPS1*SSQ) 266, 266, 270                     ARI12400
  266 CONTINUE                                                          ARI12410
      IEXIT=1                                                           ARI12420
      GO TO 280                                                         ARI12430
270   SSQ=SUMB                                                          ARI12440
      IF(NIT - MIT) 100, 2700, 2700                                     ARI12450
 2700 IEXIT=-1                                                          ARI12460
  280 SSQ=SUMB                                                          ARI12470
C                                                                       ARI12480
C                                                END ITERATION          ARI12490
C                                                                       ARI12500
284   I=0                                                               ARI12510
      CALL MATIN(D, NP, P, I, DET)                                      ARI12520
C                                                                       ARI12530
C     A IS THE NP*NP CORRELATION MATRIX OF THE PARAMETERS               ARI12540
C                                                                       ARI12550
      DO 7692 I=1,NP                                                    ARI12560
      II = I + NP*(I-1)                                                 ARI12570
7692  E(I) = DSQRT(D(II))                                               ARI12580
      DO 340 I=1,NP                                                     ARI12590
      JI = I + NP*(I-1) - 1                                             ARI12600
      IJ = I + NP*(I-2)                                                 ARI12610
      DO 340 J = I, NP                                                  ARI12620
      JI = JI + 1                                                       ARI12630
      A(JI) = D(JI) / (E(I)*E(J))                                       ARI12640
      IJ = IJ + NP                                                      ARI12650
340   A(IJ) = A(JI)                                                     ARI12660
      RETURN                                                            ARI12670
9004  IEXIT=-4                                                          ARI12680
      RETURN                                                            ARI12690
9005  IEXIT=-2                                                          ARI12700
      RETURN                                                            ARI12710
C                                                                       ARI12720
C     IEXIT=1  NO ERROR DETECTED                                        ARI12730
C     IEXIT=-1  MAXIMUM NUMBER OF ITERATIONS ATTAINED                   ARI12740
C     ON EXIT, ARRAYS TH AND W ARE AVAILABLE IF IEXIT =1 OR IEXIT=-1    ARI12750
C     IEXIT=-4  FAILURE OCCURS IN BACKFORECASTING PROCESS               ARI12760
C     IEXIT=-2,FAILURE IN REDUCTION STEP  .                             ARI12770
C     IEXIT=-3 , FAILURE IN MATRIX INVERSION ROUTINE, IS NOT            ARI12780
C                  DETECTED BY  MATIN                                   ARI12790
                                                                        ARI12800
      END                                                               ARI12810


      SUBROUTINE INIT1(PA,NP,NDR,NOB,IWBF,IDF)                          ARI00010
                                                                        ARI00020
C----      SUBROUTINE INIT1                                             ARI00030
C                                                                       ARI00040
C      SETS UP PARAMETER VALUES OF LABELED COMMON /UNIOUT/              ARI00050
C      INSERTS USER SUPPLIED INITIAL ESTIMATES TO TOTAL MODEL PARAMETER ARI00060
C                                                                       ARI00070
C----                                                                   ARI00080
C                                                                       ARI00090
C      PA:    OUTPUT. TOTAL MODEL PARAMETER ARRAY                       ARI00100
C      NP:    ON EXIT CONTAINS NUMBER OF PARAMETERS                     ARI00110
C      NDR:   ON EXIT CONTAINS NUMBER OF RESIDUALS                      ARI00120
C      NOB:   ON ENTRY CONTAINS NUMBER OF TIME SERIES OBSERVED VALUES   ARI00130
C      IWBF:  ON ENTRY CONTAINS 1 IF BACKFORECASTING IS TO BE USED,     ARI00140
C             0 OTHERWISE                                               ARI00150
C      IDF:   ON EXIT CONTAINS DEGREES OF FREEDOM                       ARI00160
C                                                                       ARI00170
C----                                                                   ARI00180
                                                                        ARI00190
      IMPLICIT REAL*8(A-H , O-Z)
      DIMENSION PA(NP)                                                  ARI00200
      COMMON/UNIIN/MFAC(3),AVEPA,ND(3),IOD(3),INC(6),IOPA(10)           ARI00210
     *,            UPA(10),ITREND,TREPA                                 ARI00220
      COMMON/UNIOUT/NAR,NMA,NDIFAC,NARMAF,MAX1,MAX11,MAX2,MBO           ARI00230
C                                                                       ARI00240
C----COMMON /UNIOUT/                                                    ARI00250
C                                                                       ARI00260
      NAR=MFAC(1)                                                       ARI00270
      NMA=MFAC(3)                                                       ARI00280
      NDIFAC=MFAC(2)                                                    ARI00290
      NARMAF=NAR+NMA                                                    ARI00300
      MAX1=0                                                            ARI00310
      IF(NDIFAC.EQ.0) GOTO 11                                           ARI00320
      DO 8 J=1,NDIFAC                                                   ARI00330
    8 MAX1=MAX1+ND(J)*IOD(J)                                            ARI00340
   11 KK=0                                                              ARI00350
      LDIF=MAX1                                                         ARI00360
      IF(NAR.EQ.0) GOTO 12                                              ARI00370
      DO 9 J=1,NAR                                                      ARI00380
      KK=KK+INC(J)                                                      ARI00390
    9 MAX1=MAX1+IOPA(KK)                                                ARI00400
   12 MAX11=MAX1+1                                                      ARI00410
      MAX2=0                                                            ARI00420
      IF(NMA.EQ.0) GOTO 13                                              ARI00430
      DO 14 J=1,NMA                                                     ARI00440
      LL=NAR+J                                                          ARI00450
      KK=KK+INC(LL)                                                     ARI00460
   14 MAX2=MAX2+IOPA(KK)                                                ARI00470
   13 MBO=MAX0(MAX1,MAX2)                                               ARI00480
C                                                                       ARI00490
C      SETS UP NUMBER OF TOTAL MODEL PARAMETERS                         ARI00500
C                                                                       ARI00510
      NP=0                                                              ARI00520
      DO 1125 J=1,NARMAF                                                ARI00530
 1125 NP=NP+INC(J)                                                      ARI00540
      IF(NDIFAC.EQ.0) NP=NP+1                                           ARI00550
      IF(ITREND.GT.0) NP=NP+1                                           ARI00560
C                                                                       ARI00570
C      SETS UP NUMBER OF MODEL RESIDUALS                                ARI00580
C                                                                       ARI00590
      NDR=NOB                                                           ARI00600
      IF(MAX1.EQ.LDIF) NDR=MIN0(NOB,NOB-LDIF+MAX2)                      ARI00610
      IF(IWBF.EQ.0) NDR=NOB-MAX1                                        ARI00620
      IDF=NOB-LDIF-NP                                                   ARI00630
      IF(IWBF.EQ.0) IDF=NDR-NP                                          ARI00640
C                                                                       ARI00650
C      INSERTS USER SUPPLIED INITIAL ESTIMATES                          ARI00660
C                                                                       ARI00670
      L=0                                                               ARI00680
      IF(NDIFAC.GT.0) GO TO 2341                                        ARI00690
      L=1                                                               ARI00700
      PA(L)=AVEPA                                                       ARI00710
 2341 CONTINUE                                                          ARI00720
      MAX=NP-L                                                          ARI00730
      IF(ITREND.GT.0) MAX=MAX-1                                         ARI00740
      IF(MAX.EQ.0) GO TO 71                                             ARI00750
      DO 45 J=1,MAX                                                     ARI00760
      L=L+1                                                             ARI00770
   45 PA(L)=UPA(J)                                                      ARI00780
   71 CONTINUE                                                          ARI00790
      IF(ITREND.LE.0) GO TO 73                                          ARI00800
      L=L+1                                                             ARI00810
      PA(L)=TREPA                                                       ARI00820
   73 CONTINUE                                                          ARI00830
C                                                                       ARI00840
      RETURN                                                            ARI00850
      END                                                               ARI00860


      SUBROUTINE PARTW(W,LW,NOB,IPA,IPZ,IPT,IPTEMP,IPC,IPCF,IPV,IEXIT)  ARI00870
C                                                                       ARI00880
C----      SUBROUTINE PARTW                                             ARI00890
C                                                                       ARI00900
C      PARTITIONING OF WORKSPACE W, TEST DIMENSION LW AND SET W TO 0    ARI00910
C      THIS PROGRAM IS TO BE USED IN CONJUNCTION WITH PROGRAM BCKF      ARI00920
C                                                                       ARI00930
C----                                                                   ARI00940
C                                                                       ARI00950
C      W:     REAL ARRAY USED AS WORKSPACE                              ARI00960
C      LW:    DIMENSION OF W AS DECLARED IN CALLING PROGRAM             ARI00970
C      NOB:   NUMBER OF OBSERVATION OF TIME SERIES                      ARI00980
C      IPA,IPZ,IPT,IPTEMP,IPC,IPCF,IPV:  THESE PARAMETERS SEPARATE      ARI00990
C                                          THE WORKING SPACE            ARI01000
C      IEXIT: IEXIT IS SET TO -1 IF THE NUMBER OF POSITIONS OF W REQUIREARI01010
C              IS GREATER THAN LW. OTHERWISE IEXIT IS SET TO 0          ARI01020
C                                                                       ARI01030
C----                                                                   ARI01040
C                                                                       ARI01050
      IMPLICIT REAL*8(A-H , O-Z)
      DIMENSION W(LW)                                                   ARI01060
      DATA IMAX/200/                                                    ARI01070
      IPA=1                                                             ARI01080
      IPZ=IPA+NOB*2+1                                                   ARI01090
      IPT=IPZ+NOB*2+1                                                   ARI01100
      IPTEMP=IPT+IMAX                                                   ARI01110
      IPC=IPTEMP+IMAX                                                   ARI01120
      IPCF=IPC+IMAX                                                     ARI01130
      IPV=IPCF+IMAX                                                     ARI01140
C                                                                       ARI01150
      JMAX=IPV+IMAX-1                                                   ARI01160
      IF(JMAX.GT.LW) GOTO 1                                             ARI01170
      IEXIT=0                                                           ARI01180
      DO 2 I=1,LW                                                       ARI01190
    2 W(I)=0.0                                                          ARI01200
      RETURN                                                            ARI01210
    1 IEXIT=-1                                                          ARI01220
      RETURN                                                            ARI01230
      END                                                               ARI01240


      SUBROUTINE DEVPA1(TH,NP)                                          ARI01250
      IMPLICIT REAL*8(A-H , O-Z)
      DIMENSION TH(NP)                                                  ARI01260
      COMMON/UNIIN/MFAC(3),AVEPA,ND(3),IOD(3),INC(6)                    ARI01270
     *,            IOPA(10),UPA(10),ITREND,TREPA                        ARI01280
      COMMON/UNIOUT/NAR,NMA,NDIFAC,NARMAF,MAX1,MAX11,MAX2,MBO           ARI01290
C                                                                       ARI01300
C      SETS UP PARAMETERS CURRENT ESTIMATES                             ARI01310
C      TO INDIVIDUAL ARRAYS                                             ARI01320
C                                                                       ARI01330
      L=0                                                               ARI01340
      IF(NDIFAC.NE.0) GOTO 2                                            ARI01350
      L=1                                                               ARI01360
      AVEPA=TH(L)                                                       ARI01370
    2 MAX=NP-L                                                          ARI01380
      IF(ITREND.NE.0) MAX=MAX-1                                         ARI01390
      IF(MAX.EQ.0) GOTO 5                                               ARI01400
      DO 6 J=1,MAX                                                      ARI01410
      L=L+1                                                             ARI01420
    6 UPA(J)=TH(L)                                                      ARI01430
    5 IF(ITREND.EQ.0) GOTO 7                                            ARI01440
      L=L+1                                                             ARI01450
      TREPA=TH(L)                                                       ARI01460
    7 CONTINUE                                                          ARI01470
C                                                                       ARI01480
      RETURN                                                            ARI01490
      END                                                               ARI01500


      REAL*8 FUNCTION RMEAN(X,NOB)                                      ARI01510
C                                                                       ARI01520
C----      DOUBLE PRECISION FUNCTION RMEAN                              ARI01530
C                                                                       ARI01540
C      CALCULATES THE SAMPLE MEAN OF A SERIES OF NOB OBSERVATIONS       ARI01550
C                                                                       ARI01560
C----                                                                   ARI01570
C                                                                       ARI01580
C      THIS PROGRAM IS TAKEN FROM " A COMPUTER PROGRAM FOR THE          ARI01590
C      ANALYSIS OF TIME SERIES MODELS USING THE BOX-JENKINS PHILOSOPHY "ARI01600
C      BY D.J.PACK (1979) AUTOMATIC FORECASTING SYSTEMS, INC.           ARI01610
C                                                                       ARI01620
C----                                                                   ARI01630
C                                                                       ARI01640
C      X:  OBSERVED TIME SERIES OF NOB VALUES                           ARI01650
C                                                                       ARI01660
C----                                                                   ARI01670
C                                                                       ARI01680
      IMPLICIT REAL*8(A-H , O-Z)
      DIMENSION X(NOB)                                                  ARI01690
      SUM=0                                                             ARI01700
      DO 10 J=1,NOB                                                     ARI01710
  10  SUM=SUM+X(J)                                                      ARI01720
      RMEAN=SUM/FLOAT(NOB)                                              ARI01730
      RETURN                                                            ARI01740
      END                                                               ARI01750


      REAL*8 FUNCTION CPROXY(X,XBAR,Y,YBAR,NOB)                         ARI01760
C                                                                       ARI01770
C----     DOUBLE PRECISION FUNCTION CPROXY                              ARI01780
C                                                                       ARI01790
C      CALCULATES SUM OF PRODUCTS OF DEVIATIONS OF TWO VARIABLES        ARI01800
C      FROM SPECIFIED CENTRAL VALUES                                    ARI01810
C                                                                       ARI01820
C----                                                                   ARI01830
C                                                                       ARI01840
C      THIS PROGRAM IS TAKEN FROM " A COMPUTER PROGRAM FOR THE          ARI01850
C      ANALYSIS OF TIME SERIES MODELS USING THE BOX-JENKINS PHILOSOPHY "ARI01860
C      BY D.J.PACK (1979) AUTOMATIC FORECASTING SYSTEMS, INC.           ARI01870
C                                                                       ARI01880
C----                                                                   ARI01890
C                                                                       ARI01900
C      X,Y:      REAL ARRAYS OF LENGTH NOB.                             ARI01910
C                ON ENTRY CONTAIN VALUES OF THE TWO VARIABLES           ARI01920
C      XBAR,YBAR:X'S AND Y'S MEAN RESPECTIVELY                          ARI01930
C                                                                       ARI01940
C----                                                                   ARI01950
C                                                                       ARI01960
      IMPLICIT REAL*8(A-H , O-Z)
      DIMENSION X(NOB),Y(NOB)                                           ARI01970
      S=0                                                               ARI01980
      DO 10 I=1,NOB                                                     ARI01990
  10  S=S+(X(I)-XBAR)*(Y(I)-YBAR)                                       ARI02000
      CPROXY=S                                                          ARI02010
      RETURN                                                            ARI02020
      END                                                               ARI02030


      SUBROUTINE ZCOR(C,NOB,NAC,R,SER,DM,DSD,IEXIT)                     ARI02040
C                                                                       ARI02050
C-----     SUBROUTINE ZCOR                                              ARI02060
C                                                                       ARI02070
C     CALCULATES AUTOCORRELATIONS AND THEIR STANDARD ERRORS             ARI02080
C                                                                       ARI02090
C-----                                                                  ARI02100
C                                                                       ARI02110
C     C     SERIES INPUT OF NOB OBS.                                    ARI02120
C     NAC   MAXIMUM LAG                                                 ARI02130
C     R     AUTOCORRELATION FUNCTION OF NAC VALUES                      ARI02140
C     SER  AUTOCOVARIANCE FUNCTION OF NAC VALUES                        ARI02150
C     DM    MEAN OF THE SERIES                                          ARI02160
C     DSD   STANDARD ERROR OF THE SERIES                                ARI02170
C     IEXIT IF NAC>= NOB IEXIT IS SET TO -1                             ARI02180
C           UNLESS THIS ONE ERROR IS DETECTED IEXIT                     ARI02190
C           CONTAINS ZERO ON EXIT                                       ARI02200
C                                                                       ARI02210
C-----                                                                  ARI02220
C                                                                       ARI02230
      IMPLICIT REAL*8(A-H , O-Z)
      DIMENSION C(NOB),R(NAC),SER(NAC)                                  ARI02240
      IEXIT=0                                                           ARI02250
      IF(NOB-NAC) 100,100,20                                            ARI02260
   20 DIV=NOB                                                           ARI02270
      DM=RMEAN(C,NOB)                                                   ARI02280
      DENOM=CPROXY(C,DM,C,DM,NOB)                                       ARI02290
      DSD=DSQRT(DENOM/DIV)                                              ARI02300
      DO 30 K=1,NAC                                                     ARI02310
      M=NOB-K                                                           ARI02320
      HNUM=CPROXY(C(1),DM,C(K+1),DM,M)                                  ARI02330
      SER(K)=HNUM/DIV                                                   ARI02340
      R(K)=HNUM/DENOM                                                   ARI02350
   30 CONTINUE                                                          ARI02360
      NAC1=NAC+1                                                        ARI02370
      DO 40 K=1,NAC                                                     ARI02380
      M=NAC1-K+1                                                        ARI02390
      R(M)=R(M-1)                                                       ARI02400
      SER(M)=SER(M-1)                                                   ARI02410
   40 CONTINUE                                                          ARI02420
      R(1)=1.0                                                          ARI02430
      SER(1)=DSD*DSD                                                    ARI02440
      RETURN                                                            ARI02450
  100 IEXIT=-1                                                          ARI02460
      RETURN                                                            ARI02470
      END                                                               ARI02480


      SUBROUTINE SMEADL(X,N,XMEAN)                                      INE13820
C     MEAN DELETION                                                     INE13850
      IMPLICIT REAL*8(A-H , O-Z)
      DIMENSION X(N)                                                    INE13870
      AN=N                                                              INE13880
      XMEAN=SUMF(X,N)/AN                                                INE13890
      DO 10 I=1,N                                                       INE13900
   10 X(I)=X(I)-XMEAN                                                   INE13910
      RETURN                                                            INE13920
      END                                                               INE14030


      REAL*8 FUNCTION SUMF(X,N)
      IMPLICIT REAL*8(A-H , O-Z)
      DIMENSION X(N)                                                    
      SUMF=0.0                                                          
      DO 10 I=1,N                                                       
   10 SUMF=SUMF+X(I)                                                    
      RETURN                                                            
      END                                                               


      SUBROUTINE LSQFUN(IEXIT,NDR,NP,PA,R,IW,LIW,W,LW)                  ARI02490
C                                                                       ARI02500
C-----     SUBROUTINE LSQFUN                                            ARI02510
C                                                                       ARI02520
C      LSQFUN MUST CALCULATE THE VECTOR OF VALUES F(X)                  ARI02530
C      AT ANY POINT X                                                   ARI02540
C                                                                       ARI02550
C-----                                                                  ARI02560
C                                                                       ARI02570
C      IEXIT: UNLESS AN ERROR IS DETECTED, IEXIT CONTAINS ZERO          ARI02580
C             ON EXIT. IF AN ERROR IS DETECTED IEXIT CONTAINS -1        ARI02590
C           ERROR IF ANY CAN BE DETECTED BY ONE OF THE ROUTINS          ARI02600
C            BCKF,MULTBP OR PARTW                                       ARI02610
C      NDR,NP: ON ENTRY, NDR SPECIFIES THE NUMBER OF RESIDUALS,         ARI02620
C              AND NP THE NUMBER OF VARIABLES                           ARI02630
C      PA : REAL ARRAY OF DIMENSION (NP). ON ENTRY, PA CONTAINS         ARI02640
C           THE POINT AT WHICH THE VALUES OF F(X) ARE REQUIRED          ARI02650
C      R :  REAL ARRAY OF DIMENSION (NDR). UNLESS IEXIT IS SET          ARI02660
C           TO -1, ON EXIT R MUST CONTAIN THE VALUE                     ARI02670
C           OF F(X) AT THE POINT PA.                                    ARI02680
C      IW(LIW),W(LW) : INTEGER AND REAL ARRAY RESPECTIVELY.             ARI02690
C                      USED AS WORKING SPACES                           ARI02700
C                                                                       ARI02710
C------                                                                 ARI02720
C                                                                       ARI02730
C      CALLS: BCKF,(CPROXY,MULTBP,NCHOSE,RMEAN),PARTW                   ARI02740
C                                                                       ARI02750
C-----                                                                  ARI02760
C                                                                       ARI02770
      IMPLICIT REAL*8(A-H , O-Z)
      DIMENSION PA(NP),R(NDR),IW(LIW),W(LW)                             ARI02780
      COMMON/UNIIN/MFAC(3),AVEPA,ND(3),IOD(3),INC(6)                    ARI02790
     *,            IOPA(10),UPA(10),ITREND ,TREPA                       ARI02800
      COMMON/UNIOUT/NAR,NMA,NDIFAC,NARMAF,MAX1,MAX11,MAX2,MBO           ARI02810
      COMMON/TSOBS/IWBF,NOB,X(5500)                                     ARI02820
      DATA MBF/101/                                                     ARI02830
      IEXIT=0                                                           ARI02840
C                                                                       ARI02850
C      REAL ARRAY R IS SET TO ZERO                                      ARI02860
C                                                                       ARI02870
      DO 1 I=1,NDR                                                      ARI02880
    1 R(I)=0.0                                                          ARI02890
C                                                                       ARI02900
C      PARAMETERS IPA,IPZ,IPT,IPTEMP,IPC,IPCF,IPV SEPARATE              ARI02910
C      THE WORKING SPACE ARRAY W                                        ARI02920
C                                                                       ARI02930
      CALL PARTW(W,LW,NOB,IPA,IPZ,IPT,IPTEMP,IPC,IPCF,IPV,IEXIT)        ARI02940
      IF(IEXIT.EQ.-1)RETURN                                             ARI02950
C                                                                       ARI02960
C      EXPANSION OF THE TERMS IN THE ARMA MODEL FOR GIVEN               ARI02970
C      PARAMETERS VALUES                                                ARI02980
C                                                                       ARI02990
      CALL BCKF(PA,NP,R,NDR,X,NOB,W(IPA),W(IPZ),W(IPT),W(IPTEMP)        ARI03000
     *,         W(IPC),W(IPCF),W(IPV),IWBF,MBF,IEXIT                    ARI03010
     +,IW,LIW,W,LW)
      RETURN                                                            ARI03020
C                                                                       ARI03030
      END                                                               ARI03040


      REAL*8 FUNCTION DOTPRD(P, X, Y)                                   ARI03050
C                                                                       ARI03060
C  ***  RETURN THE INNER PRODUCT OF THE P-VECTORS X AND Y.  ***         ARI03070
C                                                                       ARI03080
C                                                                       ARI03090
C     THIS PROGRAM IS TAKEN FROM :                                      ARI03100
C     J.E.DENNIS, D.M.GAY AND R.E.WELSCH, 'ALGORITHM 573 NL2SOL-        ARI03110
C     AN ADAPTIVE NONLINEAR LEAST-SQUARES ALGORITHM E4| ',             ARI03120
C     ACM TRANSECTIONS ON MATHEMATICAL SOFTWARE, VOL.7, NO.3,           ARI03130
C     SEPTEMBER 1981                                                    ARI03140
C                                                                       ARI03150
      IMPLICIT REAL*8(A-H , O-Z)
      INTEGER P                                                         ARI03160
      REAL*8 X(P), Y(P)                                                 ARI03170
C                                                                       ARI03180
      INTEGER I                                                         ARI03190
      REAL*8 ONE, SQTETA, T, ZERO                                       ARI03200
C/+                                                                     ARI03210
      REAL*8 ZAMAX1, ZABS                                               ARI03220
C/                                                                      ARI03230
      EXTERNAL RMDCON                                                   ARI03240
      REAL*8 RMDCON                                                     ARI03250
C                                                                       ARI03260
C  ***  RMDCON(2) RETURNS A MACHINE-DEPENDENT CONSTANT, SQTETA, WHICH   ARI03270
C  ***  IS SLIGHTLY LARGER THAN THE SMALLEST POSITIVE NUMBER THAT       ARI03280
C  ***  CAN BE SQUARED WITHOUT UNDERFLOWING.                            ARI03290
C                                                                       ARI03300
C/6                                                                     ARI03310
      DATA ONE/1.D+0/, SQTETA/0.D+0/, ZERO/0.D+0/                       ARI03320
C/7                                                                     ARI03330
C     PARAMETER (ONE=1.D+0, ZERO=0.D+0)                                 ARI03340
C     DATA SQTETA/0.D+0/                                                ARI03350
C/                                                                      ARI03360
C                                                                       ARI03370
      DOTPRD = ZERO                                                     ARI03380
      IF (P .LE. 0) GO TO 999                                           ARI03390
      IF (SQTETA .EQ. ZERO) SQTETA = RMDCON(2)                          ARI03400
      DO 20 I = 1, P                                                    ARI03410
       T = DMAX1(DABS(X(I)), DABS(Y(I)))                                ARI03420
         IF (T .GT. ONE) GO TO 10                                       ARI03430
         IF (T .LT. SQTETA) GO TO 20                                    ARI03440
         T = (X(I)/SQTETA)*Y(I)                                         ARI03450
        IF (DABS(T) .LT. SQTETA) GO TO 20                               ARI03460
 10      DOTPRD = DOTPRD + X(I)*Y(I)                                    ARI03470
 20   CONTINUE                                                          ARI03480
C                                                                       ARI03490
 999  RETURN                                                            ARI03500
C  ***  LAST CARD OF DOTPRD FOLLOWS  ***                                ARI03510
      END                                                               ARI03520


      SUBROUTINE MATIN(A, NVAR, B, NB, DET)                             ARI12820
C                                                                       ARI12830
C-----     SUBROUTINE MATIN                                             ARI12840
C                                                                       ARI12850
C     AUXILIARY ROUTINE FOR MATRIX INVERSION                            ARI12860
C                                                                       ARI12870
C-----                                                                  ARI12880
C                                                                       ARI12890
C     THIS PROGRAM IS TAKEN FROM 'A COMPUTER PROGRAM FOR THE ANALYSIS   ARI12900
C     OF TIME SERIES MODELS USING THE BOX-JENKINS PHILOSOPHY'  BY       ARI12910
C     D.J.PACK (1979) AUTOMATIC FORECASTING SYSTEMS, INC.               ARI12920
C                                                                       ARI12930
C-----                                                                  ARI12940
C     THE ROUTINE MATIN IS TO BE USED IN CONJUNCTION WITH GLM1 ROUTINE  ARI12950
C-----                                                                  ARI12960
C                                                                       ARI12970
      IMPLICIT REAL*8(A-H , O-Z)
      DIMENSION A(NVAR, 1), B(NVAR, 1)                                  ARI12980
       PIVOTM = A(1,1)                                                  ARI12990
       DET = 1.0                                                        ARI13000
      DO 550 ICOL = 1, NVAR                                             ARI13010
       PIVOT = A(ICOL, ICOL)                                            ARI13020
      PIVOTM = DMIN1(PIVOT, PIVOTM)                                     ARI13030
       DET = PIVOT * DET                                                ARI13040
C                                                                       ARI13050
C     DIVIDE PIVOT ROW BY PIVOT ELEMENT                                 ARI13060
C                                                                       ARI13070
       A(ICOL, ICOL) = 1.0                                              ARI13080
      PIVOT = DMAX1(PIVOT, 1.D-20)                                      ARI13090
       PIVOT = A(ICOL, ICOL)/PIVOT                                      ARI13100
      DO 350 L=1,NVAR                                                   ARI13110
 350   A(ICOL, L) = A(ICOL, L)*PIVOT                                    ARI13120
      IF(NB .EQ. 0) GO TO 371                                           ARI13130
      DO 370 L=1,NB                                                     ARI13140
 370   B(ICOL, L) = B(ICOL, L)*PIVOT                                    ARI13150
C                                                                       ARI13160
C     REDUCE NON-PIVOT ROWS                                             ARI13170
C                                                                       ARI13180
 371  DO 550 L1=1,NVAR                                                  ARI13190
       IF(L1 .EQ. ICOL) GO TO 550                                       ARI13200
       T = A(L1, ICOL)                                                  ARI13210
       A(L1, ICOL) = 0.                                                 ARI13220
      DO 450 L=1,NVAR                                                   ARI13230
 450   A(L1, L) = A(L1, L) - A(ICOL, L)*T                               ARI13240
      IF(NB .EQ. 0) GO TO 550                                           ARI13250
      DO 500 L=1,NB                                                     ARI13260
 500   B(L1, L) = B(L1, L)-B(ICOL,L)*T                                  ARI13270
 550  CONTINUE                                                          ARI13280
      RETURN                                                            ARI13290
      END                                                               ARI13300


      SUBROUTINE DIFFER(Z,NOB,NDIFAC,ND,IOD,C,NOBD,IEXIT)               ARI13310
C                                                                       ARI13320
C-----     SUBROUTINE DIFFER                                            ARI13330
C                                                                       ARI13340
C     CALCULATES REQUESTED DIFFERENCE OF A TIME SERIES                  ARI13350
C                                                                       ARI13360
C-----                                                                  ARI13370
C                                                                       ARI13380
C     Z     INPUT SERIES OF NOB VALUES                                  ARI13390
C     NDIFAC : COMMON/UNIOUT/                                           ARI13400
C     ND,IOD : COMMON/UNIIN/                                            ARI13410
C     C     OUTPUT SERIES AFTER DIFFERENCING OPERATION OF NOBD          ARI13420
C             VALUES                                                    ARI13430
C     IEXIT  : IF NOBD IS FOUND LESS OR EQUAL TO ZERO IEXIT IS          ARI13440
C              SET TO -1. UNLESS THIS ERROR IS DETECTED, IEXIT          ARI13450
C              CONTAINS ZERO ON EXIT.                                   ARI13460
C                                                                       ARI13470
C-----                                                                  ARI13480
C                                                                       ARI13490
         IMPLICIT REAL*8 (A-H,O-Z)                                      ARI13500
      DIMENSION Z(NOB),ND(3),IOD(3),C(NOB)                              ARI13510
      IEXIT=0                                                           ARI13520
      DO 1 J=1,NOB                                                      ARI13530
   1  C(J)=Z(J)                                                         ARI13540
      NOBD=NOB                                                          ARI13550
      IF(NDIFAC .EQ. 0)RETURN                                           ARI13560
      DO 20 I=1,NDIFAC                                                  ARI13570
      MAX=ND(I)                                                         ARI13580
      LOSE=IOD(I)                                                       ARI13590
      IF(MAX .EQ. 0)GO TO 20                                            ARI13600
      DO 10 L=1,MAX                                                     ARI13610
      NOBD=NOBD-LOSE                                                    ARI13620
      IF(NOBD .LE. 0)GO TO 300                                          ARI13630
      DO 10 J=1,NOBD                                                    ARI13640
      K=J+LOSE                                                          ARI13650
  10  C(J)=C(K)-C(J)                                                    ARI13660
  20  CONTINUE                                                          ARI13670
      RETURN                                                            ARI13680
300   IEXIT=-1                                                          ARI13690
      RETURN                                                            ARI13700
      END                                                               ARI13710

                                                                        ARI13720
      SUBROUTINE zroots(a,m,roots,polish)
C     roots of a polynomial by Laguerre's method with deflation 
      INTEGER m,MAXM
      REAL EPS
      COMPLEX a(m+1),roots(m)
      LOGICAL polish
      PARAMETER (EPS=1.e-6,MAXM=101)
CU    USES laguer
      INTEGER i,j,jj,its
      COMPLEX ad(MAXM),x,b,c
      do 11 j=1,m+1
        ad(j)=a(j)
11    continue
      do 13 j=m,1,-1
        x=cmplx(0.,0.)
        call laguer(ad,j,x,its)
        if(abs(aimag(x)).le.2.*EPS**2*abs(real(x))) x=cmplx(real(x),0.)
        roots(j)=x
        b=ad(j+1)
        do 12 jj=j,1,-1
          c=ad(jj)
          ad(jj)=b

          b=x*b+c
12      continue
13    continue
      if (polish) then
        do 14 j=1,m
          call laguer(a,m,roots(j),its)
14      continue
      endif
      do 16 j=2,m
        x=roots(j)
        do 15 i=j-1,1,-1
          if(real(roots(i)).le.real(x))goto 10
          roots(i+1)=roots(i)
15      continue
        i=0
10      roots(i+1)=x
16    continue
      return
      END


      SUBROUTINE laguer(a,m,x,its)
C     roots of a polynomial by Laguerre's method  
      INTEGER m,its,MAXIT,MR,MT
      REAL EPSS
      COMPLEX a(m+1),x
      PARAMETER (EPSS=2.e-7,MR=8,MT=10,MAXIT=MT*MR)
      INTEGER iter,j
      REAL abx,abp,abm,err,frac(MR)
      COMPLEX dx,x1,b,d,f,g,h,sq,gp,gm,g2
      SAVE frac
      DATA frac /.5,.25,.75,.13,.38,.62,.88,1./
      do 12 iter=1,MAXIT
        its=iter
        b=a(m+1)
        err=abs(b)
        d=cmplx(0.,0.)
        f=cmplx(0.,0.)
        abx=abs(x)
        do 11 j=m,1,-1
          f=x*f+d

          d=x*d+b
          b=x*b+a(j)
          err=abs(b)+abx*err
11      continue
        err=EPSS*err
        if(abs(b).le.err) then
          return
        else
          g=d/b
          g2=g*g
          h=g2-2.*f/b
          sq=sqrt((m-1)*(m*h-g2))
          gp=g+sq
          gm=g-sq
          abp=abs(gp)
          abm=abs(gm)
          if(abp.lt.abm) gp=gm
          if (max(abp,abm).gt.0.) then
            dx=m/gp
          else
            dx=exp(cmplx(log(1.+abx),float(iter)))

          endif
        endif
        x1=x-dx
        if(x.eq.x1)return
        if (mod(iter,MT).ne.0) then
          x=x1
        else
          x=x-dx*frac(iter/MT)
        endif
12    continue
c      pause 'too many iterations in laguer'
      return
      END


      REAL*8 FUNCTION V2NORM(P,X)                                       ARI13730
      IMPLICIT REAL*8(A-H , O-Z)
      INTEGER P                                                         ARI13740
      REAL*8 X(P)                                                       ARI13750
      DATA ONE/1.D+0/,ZERO/0.D+0/                                       ARI13760
      DATA SQTETA/0.D+0/                                                ARI13770
      IF(P.GT.0)GOTO 10                                                 ARI13780
        V2NORM=ZERO                                                     ARI13790
        GOTO 999                                                        ARI13800
10    DO 20 I=1,P                                                       ARI13810
        IF(X(I).NE.ZERO)GOTO 30                                         ARI13820
20    CONTINUE                                                          ARI13830
      V2NORM=ZERO                                                       ARI13840
      GOTO 999                                                          ARI13850
30    SCALE=ABS(X(I))                                                   ARI13860
      IF(I.LT.P)GOTO 40                                                 ARI13870
      V2NORM=SCALE                                                      ARI13880
      GOTO 999                                                          ARI13890
40    T=ONE                                                             ARI13900
      IF(SQTETA.EQ.ZERO)SQTETA=RMDCON(2)                                ARI13910
      J=I+1                                                             ARI13920
      DO 60 I=J,P                                                       ARI13930
      XI=ABS(X(I))                                                      ARI13940
       IF(XI.GT.SCALE)GOTO 50                                           ARI13950
        R=XI/SCALE                                                      ARI13960
        IF(R.GT.SQTETA)T=T+R*R                                          ARI13970
      GOTO 60                                                           ARI13980
50    R=SCALE/XI                                                        ARI13990
      IF(R.LE.SQTETA)R=ZERO                                             ARI14000
      T=ONE+T*R*R                                                       ARI14010
      SCALE=XI                                                          ARI14020
60        CONTINUE                                                      ARI14030
       V2NORM=SCALE*DSQRT(T)                                            ARI14040
999    RETURN                                                           ARI14050
      END                                                               ARI14060
                                                                        ARI14070

      REAL*8 FUNCTION RMDCON(K)                                         ARI14080
C                                                                       ARI14090
C  ***  RETURN MACHINE DEPENDENT CONSTANTS                 ***          ARI14100
C                                                                       ARI14110
C     THIS PROGRAM IS TAKEN FROM :                                      ARI14120
C     J.E.DENNIS, D.M.GAY AND R.E.WELSCH, 'ALGORITHM 573 NL2SOL-        ARI14130
C     AN ADAPTIVE NONLINEAR LEAST-SQUARES ALGORITHM E4| ',             ARI14140
C     ACM TRANSECTIONS ON MATHEMATICAL SOFTWARE, VOL.7, NO.3,           ARI14150
C     SEPTEMBER 1981                                                    ARI14160
C                                                                       ARI14170
C +++  COMMENTS BELOW CONTAIN DATA STATEMENTS FOR VARIOUS MACHINES.  +++ARI14180
C +++  TO CONVERT TO ANOTHER MACHINE, PLACE A C IN COLUMN 1 OF THE   +++ARI14190
C +++  DATA STATEMENT LINE(S) THAT CORRESPOND TO THE CURRENT MACHINE +++ARI14200
C +++  AND REMOVE THE C FROM COLUMN 1 OF THE DATA STATEMENT LINE(S)  +++ARI14210
C +++  THAT CORRESPOND TO THE NEW MACHINE.                           +++ARI14220
C                                                                       ARI14230
      IMPLICIT REAL*8(A-H , O-Z)
      INTEGER K                                                         ARI14240
C                                                                       ARI14250
C  ***  THE CONSTANT RETURNED DEPENDS ON K...                           ARI14260
C                                                                       ARI14270
C  ***        K = 1... SMALLEST POS. ETA SUCH THAT -ETA EXISTS.         ARI14280
C  ***        K = 2... SQUARE ROOT OF 1.001*ETA.                        ARI14290
C  ***        K = 3... UNIT ROUNDOFF = SMALLEST POS. NO. MACHEP SUCH    ARI14300
C  ***                 THAT 1 + MACHEP .GT. 1 .AND. 1 - MACHEP .LT. 1.  ARI14310
C  ***        K = 4... SQUARE ROOT OF 0.999*MACHEP.                     ARI14320
C  ***        K = 5... SQUARE ROOT OF 0.999*BIG (SEE K = 6).            ARI14330
C  ***        K = 6... LARGEST MACHINE NO. BIG SUCH THAT -BIG EXISTS.   ARI14340
C                                                                       ARI14350
      REAL*8 BIG, ETA, MACHEP                                           ARI14360
C/+                                                                     ARI14370
      REAL*8 ZSQRT                                                      ARI14380
C/                                                                      ARI14390
      REAL*8 ONE001, PT999                                              ARI14400
C                                                                       ARI14410
      DATA ONE001/1.001D0/, PT999/0.999D0/                              ARI14420
C
      DATA  BIG , ETA , MACHEP / 3.37D+38, 8.43D-37, 1.5D-5/                                                                       ARI14430
C  +++  IBM 360, IBM 370, OR XEROX  +++                                 ARI14440
C                                                                       ARI14450
C      DATA BIG/Z7FFFFFFF/, ETA/Z00100000/, MACHEP/Z3C100000/           ARI14460
C                                                                       ARI14470
C  +++  DATA GENERAL  +++                                               ARI14480
C                                                                       ARI14490
C     DATA BIG/0.7237E+76/, ETA/0.5398E-78/, MACHEP/0.9537E-06/         ARI14500
C                                                                       ARI14510
C  +++  DEC 11  +++                                                     ARI14520
C                                                                       ARI14530
C     DATA BIG/1.7E+38/, ETA/2.9388E-39/, MACHEP/1.1921E-07/            ARI14540
C                                                                       ARI14550
C  +++  HP3000  +++                                                     ARI14560
C                                                                       ARI14570
C     DATA BIG/1.1579E+77/, ETA/8.6362E-78/, MACHEP/2.3842E-07/         ARI14580
C                                                                       ARI14590
C  +++  HONEYWELL  +++                                                  ARI14600
C                                                                       ARI14610
C     DATA BIG/O376777000000/, ETA/O404400400000/,                      ARI14620
C    1     MACHEP/O716400000000/                                        ARI14630
C                                                                       ARI14640
C  +++  DEC10  +++                                                      ARI14650
C                                                                       ARI14660
C     DATA BIG/"377777777777/, ETA/"000400000021/,                      ARI14670
C    1     MACHEP/"147400000000/                                        ARI14680
C                                                                       ARI14690
C  +++  BURROUGHS  +++                                                  ARI14700
C                                                                       ARI14710
C     DATA BIG/O0777777777777777/, ETA/O1771000000000000/,              ARI14720
C    1     MACHEP/O1301000000000000/                                    ARI14730
C                                                                       ARI14740
C  +++  CONTROL DATA  +++                                               ARI14750
C                                                                       ARI14760
C     DATA BIG/37754000000000000000B/, ETA/00024000000000000000B/,      ARI14770
C    1     MACHEP/16414000000000000000B/                                ARI14780
C                                                                       ARI14790
C  +++  PRIME  +++                                                      ARI14800
C                                                                       ARI14810
C     DATA BIG/1.7E+38/, ETA/1.47E-39/, MACHEP/2.38419E-7/              ARI14820
C                                                                       ARI14830
C  +++  UNIVAC  +++                                                     ARI14840
C                                                                       ARI14850
C     DATA BIG/1.69E+38/, ETA/5.9E-39/, MACHEP/1.4901162E-8/            ARI14860
C                                                                       ARI14870
C  +++  VAX  +++                                                        ARI14880
C                                                                       ARI14890
C     DATA BIG/1.7E+38/, ETA/2.939E-39/, MACHEP/5.9604645E-08/          ARI14900
C                                                                       ARI14910
C-------------------------------  BODY  --------------------------------ARI14920
C                                                                       ARI14930
      GO TO (10, 20, 30, 40, 50, 60), K                                 ARI14940
C                                                                       ARI14950
 10   RMDCON = ETA                                                      ARI14960
      GO TO 999                                                         ARI14970
C                                                                       ARI14980
 20   RMDCON = DSQRT(ONE001*ETA)                                        ARI14990
      GO TO 999                                                         ARI15000
C                                                                       ARI15010
 30   RMDCON = MACHEP                                                   ARI15020
      GO TO 999                                                         ARI15030
C                                                                       ARI15040
 40   RMDCON = DSQRT(PT999*MACHEP)                                      ARI15050
      GO TO 999                                                         ARI15060
C                                                                       ARI15070
 50   RMDCON = DSQRT(PT999*BIG)                                         ARI15080
      GO TO 999                                                         ARI15090
C                                                                       ARI15100
 60   RMDCON = BIG                                                      ARI15110
C                                                                       ARI15120
 999  RETURN                                                            ARI15130
C  ***  LAST CARD OF RMDCON FOLLOWS  ***                                ARI15140
      END                                                               ARI15150
                                                                        ARI15160

      SUBROUTINE BCKF(TH,NP,R,NDR,X,NOB,A,Z,T,TEMP,C,CF,V,              ARI15170
     *                 IWBF,MIT,IEXIT,IW,LIW,W,LW)                      ARI15180
C                                                                       ARI15190
C----      SUBROUTINE BCKF                                              ARI15200
C                                                                       ARI15210
C      EXPANSION OF THE TERMS IN AN AUTOREGRESSIVE-MOVING               ARI15220
C      AVERAGE MODEL FOR GIVEN PARAMETERS VALUES TO FIND                ARI15230
C      THE CORRESPONDING MODEL RESIDUALS.                               ARI15240
C                                                                       ARI15250
C----                                                                   ARI15260
C                                                                       ARI15270
C      THIS PROGRAM IS AN ADAPTATION OF ROUTINE TSMOD FROM              ARI15280
C      "A COMPUTER PROGRAM FOR THE ANALYSIS OF TIME SERIES              ARI15290
C      MODELS USING THE BOX-JENKINS PHILOSOPHY" BY                      ARI15300
C      D.J.PACK (1979) AUTOMATIC FORECASTING SYSTEM, INC.               ARI15310
C                                                                       ARI15320
C----                                                                   ARI15330
C                                                                       ARI15340
C      TH:  REAL ARRAY OF DIMENSION (NP) CONTAINING THE                 ARI15350
C           CO-ORDINATES OF THE CURRENT TOTAL MODEL                     ARI15360
C           PARAMETER VECTOR                                            ARI15370
C      R:   REAL ARRAY OF DIMENSION (NDR) . ON EXIT CONTAINS            ARI15380
C           MODEL RESIDUALS                                             ARI15390
C      X:   REAL ARRAY OF DIMENSION (NOB) . ON INPUT, X CONTAINS        ARI15400
C           ORIGINAL DATA OF THE OBSERVED TIME SERIES.                  ARI15410
C            UNCHANGED ON EXIT                                          ARI15420
C      A,Z: REAL ARRAYS OF DIMENSION AT LEAST (NOB*2+1).                ARI15430
C           USED AS WORKSPACE.                                          ARI15440
C      T,TEMP,C,CF,V: REAL ARRAYS OF DIMENSION AT LEAST                 ARI15450
C           200.USED AS WORKSPACES                                      ARI15460
C      IWBF: ON INPUT, IWBF IS SET TO 0 IF BACKFORECASTING              ARI15470
C           IS NOT BE USED, IS SET TO 1 IF BACKFORECASTING              ARI15480
C           IS TO BE USED. UNCHANGED ON EXIT.                           ARI15490
C      MIT: MAX NUMBER OF ITERATIONS TO BE PERFORMED IN                 ARI15500
C           BACKFORECASTING PROCEDURE.                                  ARI15510
C      IEXIT: UNLESS THE ROUTINE DETECTS AN ERROR, IEXIT                ARI15520
C           CONTAINS ZERO ON EXIT                                       ARI15530
C                                                                       ARI15540
C----                                                                   ARI15550
C                                                                       ARI15560
C      CALLS: CPROXY, MULTBP, NCHOSE, RMEAN, RMDCON                     ARI15570
C                                                                       ARI15580
C----                                                                   ARI15590
C                                                                       ARI15600
      IMPLICIT REAL*8(A-H , O-Z)
      DIMENSION TH(NP),R(NDR),X(1000),A(LW),Z(LW),T(LW),TEMP(LW)        ARI15610
     *,          C(LW),CF(LW),V(LW),IW(LIW),W(LW)                       ARI15620
      COMMON/UNIIN/MFAC(3),AVEPA,ND(3),IOD(3),INC(6)                    ARI15630
     *,            IOPA(10),UPA(10),ITREND,TREPA                        ARI15640
      COMMON/UNIOUT/NAR,NMA,NDIFAC,NARMAF,MAX1,MAX11,MAX2,MBO           ARI15650
      DATA TOL1/1.D-03/,TOL2/1.1D0/,TOL3/0.0005D0/                      ARI15660
      HUO=1D0                                                           ARI15670
      IEXIT=0                                                           ARI15680
      BIG=RMDCON(5)                                                     ARI15690
C                                                                       ARI15700
C      SETS UP PARAMETERS CURRENT ESTIMATES                             ARI15710
C      TO INDIVIDUAL ARRAYS                                             ARI15720
C                                                                       ARI15730
      L=0                                                               ARI15740
      IF(NDIFAC.NE.0) GOTO 2                                            ARI15750
      L=1                                                               ARI15760
      AVEPA=TH(L)                                                       ARI15770
    2 MAX=NP-L                                                          ARI15780
      IF(ITREND.NE.0) MAX=MAX-1                                         ARI15790
      IF(MAX.EQ.0) GOTO 5                                               ARI15800
      DO 6 J=1,MAX                                                      ARI15810
      L=L+1                                                             ARI15820
    6 UPA(J)=TH(L)                                                      ARI15830
    5 IF(ITREND.EQ.0) GOTO 7                                            ARI15840
      L=L+1                                                             ARI15850
      TREPA=TH(L)                                                       ARI15860
    7 CONTINUE                                                          ARI15870
C                                                                       ARI15880
C      EVALUATION OF POLYNOMIAL COEFFICIENTS C                          ARI15890
C      TO CALCULATE REQUESTED TIME SERIES DIFFERENCE                    ARI15900
C                                                                       ARI15910
      COP=1.0                                                           ARI15920
      IF(MBO .EQ. 0) GO TO 10                                           ARI15930
      DO 3 J=1,MBO                                                      ARI15940
      C(J)=0                                                            ARI15950
   3  CF(J)=0                                                           ARI15960
   10 CONTINUE                                                          ARI15970
      LDIF=0                                                            ARI15980
      IF(NDIFAC .EQ. 0)GO TO 33                                         ARI15990
      DO 30 L=1,NDIFAC                                                  ARI16000
      KJ=IOD(L)                                                         ARI16010
      LN=ND(L)                                                          ARI16020
      MAX=KJ*LN                                                         ARI16030
      DO 18 K=1,MAX                                                     ARI16040
  18  T(K)=0                                                            ARI16050
      DO 20 K=KJ,MAX,KJ                                                 ARI16060
      J=K/KJ                                                            ARI16070
       JHK1=NCHOSE(LN,J)                                                ARI16080
       JHK2=J+1                                                         ARI16090
       JHK3=(-1)**JHK2                                                  ARI16100
       T(K)=JHK3*JHK1                                                   ARI16110
   20 CONTINUE                                                          ARI16120
  30  CALL MULTBP(HUO,T,MAX,COP,TEMP,LDIF,V,IEXIT,IW,LIW,W,LW)          ARI16130
  33  CONTINUE                                                          ARI16140
  36  IF(NDIFAC .EQ. 0)GO TO 43                                         ARI16150
      MAX=LDIF                                                          ARI16160
      DO 38 K=1,MAX                                                     ARI16170
  38  C(K)=V(K)                                                         ARI16180
  43  CONTINUE                                                          ARI16190
C                                                                       ARI16200
C      CALCULATES:                                                      ARI16210
C      CRMP (=1.0) ,C(1),C(2),...=                                      ARI16220
C      FULL POLYNOMIAL EXPANSION OF AUTOREGRESSIVE INTEGRATED TERM      ARI16230
C      COP, CF(1),CF(2),... =                                           ARI16240
C      FULL POLYNOMIAL EXPANSION OF MOVING AVERAGE TERM                 ARI16250
C                                                                       ARI16260
      IF(NARMAF .EQ. 0) GO TO 68                                        ARI16270
      MIN=1                                                             ARI16280
      MAX=0                                                             ARI16290
      LOP=0                                                             ARI16300
      IF(NDIFAC .NE. 0)LOP=LDIF                                         ARI16310
      LOP1=0                                                            ARI16320
      DO 65 L=1,NARMAF                                                  ARI16330
      LN=L-NAR                                                          ARI16340
      MAX=MAX+INC(L)                                                    ARI16350
      LMULTP=IOPA(MAX)                                                  ARI16360
      DO 50 K=1,LMULTP                                                  ARI16370
  50  T(K)=0                                                            ARI16380
      DO 54 M=MIN,MAX                                                   ARI16390
      K=IOPA(M)                                                         ARI16400
  54  T(K)=UPA(M)                                                       ARI16410
      IF(LN .GT. 0)GO TO 60                                             ARI16420
  58  CALL MULTBP(HUO,T,LMULTP,COP,C,LOP,TEMP,IEXIT,IW,LIW,W,LW)        ARI16430
      GO TO 65                                                          ARI16440
  60  CALL MULTBP(HUO,T,LMULTP,COP,CF,LOP1,TEMP,IEXIT,IW,LIW,W,LW)      ARI16450
  65  MIN=MIN+INC(L)                                                    ARI16460
   68 CONTINUE                                                          ARI16470
      IF(IEXIT.EQ.-1)RETURN                                             ARI16480
C                                                                       ARI16490
C      CALCULATES MODEL RESIDUALS R(1),...,R(NDR)                       ARI16500
C                                                                       ARI16510
      PMEAN=0                                                           ARI16520
      IF(NDIFAC .EQ. 0)PMEAN=AVEPA                                      ARI16530
      DO 70 J=1,NOB                                                     ARI16540
  70  Z(J)=X(J)-PMEAN                                                   ARI16550
      CONST=0                                                           ARI16560
      IF(ITREND .NE. 0)CONST=TREPA                                      ARI16570
      MAX=NOB                                                           ARI16580
      IF(IWBF .EQ. 0)GO TO 83                                           ARI16590
C                                                                       ARI16600
C----    BACK FORECASTING    ............................               ARI16610
C                                                                       ARI16620
      MIN=NOB*2+1                                                       ARI16630
      RMAX1=MAX1                                                        ARI16640
      LOOP=0                                                            ARI16650
      NFZ=0                                                             ARI16660
  78  LOOP=LOOP+1                                                       ARI16670
      IF(LOOP .EQ. MIT)GO TO 145                                        ARI16680
      MAXH=MAX/2                                                        ARI16690
C                                                                       ARI16700
C      INTERCHANGING ORDER OF OBSERVED TIME SERIES VALUES TO INPUT      ARI16710
C      BACKWARD MODEL                                                   ARI16720
C                                                                       ARI16730
      DO 80 J=1,MAXH                                                    ARI16740
      K=MAX-J+1                                                         ARI16750
      STOR=Z(K)                                                         ARI16760
      Z(K)=Z(J)                                                         ARI16770
  80  Z(J)=STOR                                                         ARI16780
      CONST=-CONST                                                      ARI16790
C----                       ............................                ARI16800
  83  IF(MAX1 .EQ. 0)GO TO 88                                           ARI16810
      DO 86 J=1,MAX1                                                    ARI16820
  86  A(J)=0                                                            ARI16830
  88  DO 92 K=MAX11,MAX                                                 ARI16840
      A(K)=Z(K)-CONST                                                   ARI16850
C                                                                       ARI16860
C      LN IS SET TO MIN0(K-1,MBO) TO AVOID KJ = K-J <= 0                ARI16870
C      THIS POSITION HOLDS ONLY WHEN MAX1 (=P+D) < MAX2 (=Q), THAT IS   ARI16880
C      WHEN MBO = MAX2 (=Q)                                             ARI16890
C                                                                       ARI16900
      LN=MIN0(K-1,MBO)                                                  ARI16910
      IF(LN .EQ. 0)GO TO 92                                             ARI16920
      DO 90 J=1,LN                                                      ARI16930
      KJ=K-J                                                            ARI16940
  90  A(K)=A(K)-C(J)*Z(KJ)+CF(J)*A(KJ)                                  ARI16950
  92  CONTINUE                                                          ARI16960
C                                                                       ARI16970
C      END CONDITIONAL RESIDUALS CALCULATION                            ARI16980
C      (IWBF=0)                                                         ARI16990
C                                                                       ARI17000
      IF(MBO .EQ. 0 .OR. IWBF .EQ. 0)GO TO 120                          ARI17010
C                                                                       ARI17020
C     TEST IF SOME RESIDUAL VALUE IS TOO LARGE                          ARI17030
C                                                                       ARI17040
      DO 99 I=1,NOB                                                     ARI17050
      KI=I+NFZ                                                          ARI17060
      IF(A(KI).GT.BIG) GOTO 145                                         ARI17070
   99 CONTINUE                                                          ARI17080
      NUM=MAX-MAX1                                                      ARI17090
      ABAR=RMEAN(A(MAX11),NUM)                                          ARI17100
      SIG=CPROXY(A(MAX11),ABAR,A(MAX11),ABAR,NUM)/FLOAT(NUM)            ARI17110
      SIG=DSQRT(SIG)*.10                                                ARI17120
      DO 93 J=1,NOB                                                     ARI17130
      KJ=J+NFZ                                                          ARI17140
      Z(J)=Z(KJ)                                                        ARI17150
  93  A(J)=A(KJ)                                                        ARI17160
      IF(((LOOP/2)*2) .NE. LOOP)GO TO 96                                ARI17170
      SSQ=0                                                             ARI17180
      DO 94 J=1,NOB                                                     ARI17190
  94  SSQ=SSQ+A(J)*A(J)                                                 ARI17200
      IF(LOOP .EQ. 2)GO TO 95                                           ARI17210
      IF(DABS((SSQ-SSQL)/SSQL) .LT. TOL1)GO TO 120                      ARI17220
      IF((SSQ/SSQL).GT.TOL2)GO TO 145                                   ARI17230
C                                                                       ARI17240
C      TEST RESIDUALS SUM OF SQUARES AFTER BACKFORECASTING IS PERFORMED.ARI17250
C      IF(SSQ-SSQL)/SSQL<TOL1 THEN CONVERGENCE IS ATTAINED              ARI17260
C      IF TOL1< (SSQ-SSQL)/SSQL < TOL2 THEN ITERATIVE PROCEDURE         ARI17270
C      IS TO BE CONTINUED UNTIL MIT NUMBER OF ITERATIONS OR             ARI17280
C      CONVERGENCE IS ATTAINED                                          ARI17290
C      IF (SSQ-SSQL)/SSQL > TOL2 THEN CONVERGENCE FAILURE OCCURS        ARI17300
C      IN BACKFORECASTING PROCESS                                       ARI17310
C      IEXIT IS SET TO -1 IF MIT NUMBER OF ITERATIONS IS ATTAINED OR    ARI17320
C      CONVERGENCE FAILURE OCCURS                                       ARI17330
C                                                                       ARI17340
  95  SSQL=SSQ                                                          ARI17350
  96  LN=0                                                              ARI17360
      KMAX=NOB*2                                                        ARI17370
      SUM=0                                                             ARI17380
      K=NOB                                                             ARI17390
  98  K=K+1                                                             ARI17400
      IF(K .EQ. MIN)GO TO 101                                           ARI17410
      A(K)=0                                                            ARI17420
      Z(K)=CONST                                                        ARI17430
C                                                                       ARI17440
C      BACKFORECASTING OF INITIAL UNKNOWN STARTING VALUES               ARI17450
C      OF THE OBSERVED TIME SERIES                                      ARI17460
C                                                                       ARI17470
      DO 100 J=1,MBO                                                    ARI17480
      KJ=K-J                                                            ARI17490
 100  Z(K)=Z(K)+C(J)*Z(KJ)-CF(J)*A(KJ)                                  ARI17500
      IF(LN .EQ. 0)GO TO 102                                            ARI17510
      IF(K .LT. KMAX)GO TO 98                                           ARI17520
 101  NFZ=KMAX-NOB                                                      ARI17530
      MAX=KMAX                                                          ARI17540
      GO TO 78                                                          ARI17550
 102  STOR=Z(K)-CONST                                                   ARI17560
      M=LDIF                                                            ARI17570
C                                                                       ARI17580
C      DIFFERENCING BACKFORECASTED VALUE OF OBSERVED TIME SERIES        ARI17590
C                                                                       ARI17600
      IF(M.EQ.0)GO TO 113                                               ARI17610
      DO 112 J=1,M                                                      ARI17620
      KJ=K-J                                                            ARI17630
 112  STOR=STOR-V(J)*Z(KJ)                                              ARI17640
 113  IF(DABS(STOR) .LT. TOL3)GO TO 114                                 ARI17650
C                                                                       ARI17660
C      TEST IF ACTUAL BACKFORECASTED VALUE OF OBSERVED TIME SERIES      ARI17670
C      IS PRACTICALLY EQUAL TO ZERO AFTER DIFFERENCING                  ARI17680
C                                                                       ARI17690
      IF(NAR.EQ.0)GO TO 98                                              ARI17700
      JJ=K-NOB                                                          ARI17710
      R(JJ)=DABS(STOR)/RMAX1                                            ARI17720
      SUM=SUM+R(JJ)                                                     ARI17730
       IF(JJ.LT.MAX1)      GO TO 98                                     ARI17740
      LL=JJ-MAX1                                                        ARI17750
      IF(LL.GT.0)SUM=SUM-R(LL)                                          ARI17760
      IF(SUM.LT.SIG)GO TO 114                                           ARI17770
C                                                                       ARI17780
C      TEST IF MEAN OF THE ABSOLUTE VALUE OF BACKFORECASTED VALUES OF   ARI17790
C      OBSERVED TIME SERIES AFTER DIFFERENCING IS LESS THAN 0.1 TIMES   ARI17800
C      STANDARD ERROR OF RESIDUALS                                      ARI17810
C                                                                       ARI17820
      GO TO 98                                                          ARI17830
 114  LN=1                                                              ARI17840
      KMAX=MIN0(KMAX,K+MAX1-LDIF-1)                                     ARI17850
      IF((KMAX-K) .LE. 0)GO TO 101                                      ARI17860
      GO TO 98                                                          ARI17870
C                                                                       ARI17880
C      END NUMBER 'LOOP' BACKFORECASTING ITERATION                      ARI17890
C                                                                       ARI17900
 120  K=NOB-NDR                                                         ARI17910
      DO 135 J=1,NDR                                                    ARI17920
      LN=J+K                                                            ARI17930
 135  R(J)=A(LN)                                                        ARI17940
      RETURN                                                            ARI17950
 145  IEXIT=-1                                                          ARI17960
C                                                                       ARI17970
C      CONVERGENCE FAILURE IN BACKFORECASTING PROCESS                   ARI17980
C      OR MIT ITERATIONS PERFORMED                                      ARI17990
C                                                                       ARI18000
      RETURN                                                            ARI18010
      END                                                               ARI18020


      FUNCTION NCHOSE(N,K)                                              ARI18030
C                                                                       ARI18040
C----      SUBROUTINE NCHOSE                                            ARI18050
C                                                                       ARI18060
C      AUXILIARY SUBROUTINE TO CALCULATE NUMBER OF COMBINATIONS         ARI18070
C      OF N THINGS TAKEN K AT A TIME                                    ARI18080
C                                                                       ARI18090
C----                                                                   ARI18100
C                                                                       ARI18110
C      THIS PROGRAM IS TAKEN FROM " A COMPUTER PROGRAM FOR THE          ARI18120
C      ANALYSIS OF TIME SERIES MODELS USING THE BOX-JENKINS PHILOSOPHY "ARI18130
C      BY D.J.PACK (1979) AUTOMATIC FORECASTING SYSTEMS, INC.           ARI18140
C                                                                       ARI18150
C----                                                                   ARI18160
C                                                                       ARI18170
      IF(K)200,4,6                                                      ARI18180
   4  NCHOSE=1                                                          ARI18190
      RETURN                                                            ARI18200
   6  IF(N-K)200,4,8                                                    ARI18210
   8  KK=MIN0(K,N-K)                                                    ARI18220
      NN=1                                                              ARI18230
      DO 10 I=1,KK                                                      ARI18240
  10  NN=(NN*(N-I+1))/I                                                 ARI18250
      NCHOSE=NN                                                         ARI18260
      RETURN                                                            ARI18270
  200 CONTINUE                                                          ARI18280
C                                                                       ARI18290
C      NCHOSE UNDEFINED. RESULT IS SET TO ZERO                          ARI18300
C                                                                       ARI18310
      NCHOSE=0                                                          ARI18320
      RETURN                                                            ARI18330
      END                                                               ARI18340
                                                                        ARI18350

      SUBROUTINE MULTBP(CRMP,RMP,LRMP,COP,OP,LOP,TEMP,IEXIT             ARI18360
     +,IW,LIW,W,LW)
C                                                                       ARI18370
C----      SUBROUTINE MULTBP                                            ARI18380
C                                                                       ARI18390
C      SOLVES FOR COEFFICIENTS T(0),T(1),T(2),... IN THE POLYNOMIAL     ARI18400
C      EQUATION  ( R(0) - R(1)*B - R(2)*B^2 - ...) *                    ARI18410
C      * ( P(0) - P(1)*B - P(2)*B^2 - ... ) = ( T(0) - T(1)*B - ...)    ARI18420
C      BY MULTIPLYING FACTORS ON LEFT SIDE  OF EQUATION.                ARI18430
C      R(I)'S AND P(J)'S ASSUMED GIVEN.                                 ARI18440
C                                                                       ARI18450
C----                                                                   ARI18460
C                                                                       ARI18470
C      THIS PROGRAM IS TAKEN FROM " A COMPUTER PROGRAM FOR THE          ARI18480
C      ANALYSIS OF TIME SERIES MODELS USING THE BOX-JENKINS PHILOSOPHY "ARI18490
C      BY D.J.PACK (1979) AUTOMATIC FORECASTING SYSTEMS, INC.           ARI18500
C                                                                       ARI18510
C----                                                                   ARI18520
C                                                                       ARI18530
C      CRMP:  R(0) COEFFICIENT                                          ARI18540
C      RMP:   R VECTOR OF B^I OPERATORS COEFFICIENTS ON INPUT           ARI18550
C      LRMP:  NUMBER OF INPUT COEFFICIENTS R(1),R(2),...                ARI18560
C      COP:   P(0) COEFFICIENT. ON OUTPUT, COP IS SET TO T(0)           ARI18570
C             COEFFICIENT                                               ARI18580
C      OP:    P VECTOR OF B^J OPERATORS COEFFICIENTS ON INPUT.          ARI18590
C             ON OUTPUT OP IS SET TO T OUTPUT VECTOR                    ARI18600
C      LOP:   NUMBER OF INPUT COEFFICIENTS P(1),P(2),....               ARI18610
C             ON EXIT LOP IS SET TO THE NUMBER OF COEFFICIENTS T(1),T(2)ARI18620
C      TEMP:  REAL ARRAY USED AS WORKING SPACE                          ARI18630
C      IEXIT:  BEFORE ENTRY, IEXIT MUST BE ASSIGNED A VALUE. UNCHANGED  ARI18640
C              ON EXIT UNLESS THE ROUTINE DETECTS AN ERROR. IF SO,      ARI18650
C              IEXIT IS SET TO -1.                                      ARI18660
C                                                                       ARI18670
C----                                                                   ARI18680
C                                                                       ARI18690
      IMPLICIT REAL*8(A-H , O-Z)
      DIMENSION RMP(LW),OP(LW),TEMP(LW),IW(LIW),W(LW)                   ARI18700
      IF( LRMP  .LT. 0 .OR. LOP .LT. 0) GO TO 300                       ARI18710
      LTEMP=LRMP+LOP                                                    ARI18720
      IF(LTEMP .EQ. 0)GO TO 40                                          ARI18730
      DO 15 J=1,LTEMP                                                   ARI18740
      TEMP(J)=0                                                         ARI18750
  15  IF(J .LE. LOP)TEMP(J)=OP(J)*CRMP                                  ARI18760
      IF( LRMP  .EQ. 0) GO TO 30                                        ARI18770
      DO 25 J=1,LRMP                                                    ARI18780
      TEMP(J)=TEMP(J)+RMP(J)*COP                                        ARI18790
      IF(LOP .EQ. 0) GO TO 25                                           ARI18800
      DO 20 I=1,LOP                                                     ARI18810
      JI=J+I                                                            ARI18820
  20  TEMP(JI)=TEMP(JI)-OP(I)*RMP(J)                                    ARI18830
  25  CONTINUE                                                          ARI18840
  30  DO 35 J=1,LTEMP                                                   ARI18850
  35  OP(J)=TEMP(J)                                                     ARI18860
  40  COP=CRMP*COP                                                      ARI18870
      LOP=LTEMP                                                         ARI18880
      RETURN                                                            ARI18890
  300 CONTINUE                                                          ARI18900
C                                                                       ARI18910
C      ALMOST ONE OF THE MAXIMUM B POWERS OF TWO POLYNOMIALS TO BE      ARI18920
C      MULTIPLIED IS SET TO A NEGATIVE NUMBER ON ENTRY                  ARI18930
C                                                                       ARI18940
      IEXIT=-1                                                          ARI18950
      RETURN                                                            ARI18960
      END                                                               ARI18970


      SUBROUTINE FLIKAM(P,MP,Q,MQ,W,E,N,SUMSQ,FACT,VW,VL,
     *                  MRP1,VK,MR,TOLER,IFAULT)
      IMPLICIT DOUBLE PRECISION(A-H,O-Z)
C
C     ALGORITHM AS 197  APPLIED STATISTICS (1984) VOL.33, NO.1
C
C     COMPUTES THE LIKELIHOOD FUNCTION OF AN AUTOREGRESSIVE-MOVING
C     AVERAGE PROCESS, EXPRESSED AS FACT * SUMSQ.
C
C     3/96 CAC  FROM STATLIB; CONVERTED TO DOUBLE PRECISION, NO TABS,
C     UPPERCASE, MIN --> MIN0, MAX --> MAX0 TO AGREE WITH JOURNAL
C     LISTING.  FIXED 4 BUGS (SEE COMMENTS WITH ORIGINAL BELOW).
C
C     TESTED SUCCESSFULLY ON THE FOLLOWING MODELS AND BENCHMARK RESULTS:
C      MA(1)  BOX-JENKINS(1976) SERIES A  -  ARIMA(0,1,1).
C      MA(2)  OSBORN(1976) USING 28-OBS SECTIONS OF B-J SERIES C
C      ARMA(1,1)  BOX-JENKINS(1976) SERIES A  (MEAN REMOVED FIRST).
C      ARMA(2,1)  DAVID CUSHMAN'S SERIES.  COMPARED WITH TSP PROC CODE
C                 THAT IMPLEMENTED NEWBOLD(1974) METHOD.
C      ARMA(3,1), ARMA(4,1), ARMA(5,1)   COMPARED WITH GAUSS RESULTS USING
C                 ANSLEY(1979) METHOD, WHICH IS NOT COMPLETELY EXACT ML
C                 (IS MISSING THE QUADRATIC FORM IN THE INITIAL RESIDUALS
C                  MENTIONED IN JUDGE, ET AL (1980) THEORY AND PRACTICE).
C
C     ORIGINAL CODE FROM STATLIB USED P(MP),Q(MQ) BELOW, BUT THIS
C     FAILS ON COMPUTERS LIKE VAX/VMS WHEN MP OR MQ ARE ZERO.
      DIMENSION P(MP),Q(MQ),W(N),E(N),VW(MRP1),VL(MRP1),VK(MR)
C
      DATA EPSIL1 /1.D-10/
      DATA ZERO, P0625, ONE, TWO, FOUR, SIXTEN /
     *   0.D0, 0.0625D0, 1.D0, 2.D0, 4.D0, 16.D0/
C
      FACT = ZERO
      DETMAN = ONE
      DETCAR = ZERO
      SUMSQ = ZERO
      MXPQ = MAX0(MP, MQ)
      MXPQP1 = MXPQ + 1
      MQP1 = MQ + 1
      MPP1 = MP + 1
C
C      CALCULATION OF THE AUTOCOVARIANCE FUNCTION OF A PROCESS WITH
C      UNIT INNOVATION VARIANCE (VW) AND THE COVARIANCE BETWEEN THE
C      VARIABLE AND THE LAGGED INNOVATIONS (VL).
C
      CALL TWACF(P,MP,Q,MQ,VW,MXPQP1,VL,MXPQP1,VK,MXPQ,IFAULT)
      IF (MR .NE. MAX0(MP, MQP1)) IFAULT = 6
      IF (MRP1 .NE. MR + 1) IFAULT = 7
      IF (IFAULT .GT. 0) RETURN
C
C      COMPUTATION OF THE FIRST COLUMN OF MATRIX P (VK)
C
      VK(1) = VW(1)
      IF (MR .EQ. 1) GO TO 150
      DO 140 K = 2, MR
        VK(K) = ZERO
        IF (K .GT. MP) GO TO 120
        DO 110 J = K, MP
          JP2MK = J + 2 - K
C          ORIGINAL CODE FROM STATLIB
C          VK(K) = VK(K) + P(J) + VW(JP2MK)
C          FIXED TO AGREE WITH JOURNAL LISTING AND EQN (10.2)
          VK(K) = VK(K) + P(J) * VW(JP2MK)
  110   CONTINUE
  120   IF (K .GT. MQP1) GO TO 140
        DO 130 J = K, MQP1
          JP1MK = J + 1 - K
          VK(K) = VK(K) - Q(J-1) * VL(JP1MK)
  130   CONTINUE
  140 CONTINUE
C
C      COMPUTATION OF THE INITIAL VECTORS L AND K (VL, VK).
C
  150 R = VK(1)
      VL(MR) = ZERO
      DO 160 J = 1, MR
        VW(J) = ZERO
        IF (J .NE. MR) VL(J) = VK(J+1)
        IF (J .LE. MP) VL(J) = VL(J) + P(J) * R
        VK(J) = VL(J)
  160 CONTINUE
C
C      INITIALIZATION
C
      LAST = MPP1 - MQ
      LOOP = MP
      JFROM = MPP1
      VW(MPP1) = ZERO
      VL(MXPQP1) = ZERO
C
C      EXIT IF NO OBSERVATION, ELSE LOOP ON TIME.
C
      IF (N .LE. 0) GO TO 500
      DO 290 I = 1, N
C
C      TEST FOR SKIPPED UPDATING
C
        IF (I .NE. LAST) GO TO 170
        LOOP = MIN0(MP, MQ)
        JFROM = LOOP + 1
C
C      TEST FOR SWITCHING
C
        IF (MQ .LE. 0) GO TO 300
  170   IF (R .LE. EPSIL1) GO TO 400
        IF (DABS(R - ONE) .LT. TOLER .AND. I .GT. MXPQ) GO TO 300
C
C      UPDATING SCALARS
C
        DETMAN = DETMAN * R
  190   IF (DABS(DETMAN) .LT. ONE) GO TO 200
        DETMAN = DETMAN * P0625
        DETCAR = DETCAR + FOUR
        GO TO 190
  200   IF (DABS(DETMAN) .GE. P0625) GO TO 210
        DETMAN = DETMAN * SIXTEN
        DETCAR = DETCAR - FOUR
        GO TO 200
  210   VW1 = VW(1)
        A = W(I) - VW1
        E(I) = A / DSQRT(R)
        AOR = A / R
        SUMSQ = SUMSQ + A * AOR
        VL1 = VL(1)
        ALF = VL1 / R
        R = R - ALF * VL1
        IF (LOOP .EQ. 0) GO TO 230
C
C      UPDATING VECTORS
C
        DO 220 J = 1, LOOP
          FLJ = VL(J+1) + P(J) * VL1
          VW(J) = VW(J+1) + P(J) * VW1 + AOR * VK(J)
          VL(J) = FLJ - ALF * VK(J)
          VK(J) = VK(J) - ALF * FLJ
  220   CONTINUE
  230   IF (JFROM .GT. MQ) GO TO 250
        DO 240 J = JFROM, MQ
          VW(J) = VW(J+1) + AOR * VK(J)
          VL(J) = VL(J+1) - ALF * VK(J)
          VK(J) = VK(J) - ALF * VL(J+1)
  240   CONTINUE
  250   IF (JFROM .GT. MP) GO TO 270
        DO 260 J = JFROM, MP
  260   VW(J) = VW(J+1) + P(J) * W(I)
  270   CONTINUE
  290 CONTINUE
      GO TO 390
C
C      QUICK RECURSIONS
C
  300 NEXTI = I
      IFAULT = -NEXTI
      DO 310 I = NEXTI, N
  310 E(I) = W(I)
      IF (MP .EQ. 0) GO TO 340
      DO 330 I = NEXTI, N
        DO 320 J = 1, MP
          IMJ = I - J
          E(I) = E(I) - P(J) + W(IMJ)
  320   CONTINUE
  330 CONTINUE
  340 IF (MQ .EQ. 0) GO TO 370
      DO 360 I = NEXTI, N
        DO 350 J = 1, MQ
          IMJ = I - J
          E(I) = E(I) + Q(J) * E(IMJ)
  350   CONTINUE
  360 CONTINUE
C
C      RETURN SUM OF SQUARES AND DETERMINANT
C
  370 DO 380 I = NEXTI, N
  380 SUMSQ = SUMSQ + E(I) * E(I)
C
C      CODE FOR CONDITIONAL SUM OF SQUARES.
C      REPLACES ALL EXECUTABLE UP TO AND INCLUDING THAT LABELLED 380.
C
C            FACT = ZERO
C            DETMAN = ONE
C            DETCAR = ZERO
C            SUMSQ = ZERO
C            MXPQ = MAX0(MP, MQ)
C            DO 380 I = MXPQ, N
C              E(I) = W(I)
C              IF (MP .LE. 0) GO TO 340
C              DO 320 J = 1, MP
C                IMJ = I - J
C                E(I) = E(I) - P(J) * W(IMJ)
C        320        CONTINUE
C        340        IF (MQ .LE. 0) GO TO 380
C              DO 350 J = 1, MQ
C                IMJ = I - J
C                E(I) = E(I) + Q(J) * E(IMJ)
C        350        CONTINUE
C        380      SUMSQ = SUMSQ + E(I) * E(I)
C
  390 FN = N
      FACT = DETMAN ** (ONE / FN) * TWO ** (DETCAR / FN)
      RETURN
C
C      EXECUTION ERRORS
C
  400 IFAULT = 8
      RETURN
  500 IFAULT = 9
      RETURN
      END
C-----------------------
      SUBROUTINE TWACF(P,MP,Q,MQ,ACF,MA,CVLI,MXPQP1,ALPHA,MXPQ,IFAULT)
      IMPLICIT DOUBLE PRECISION(A-H,O-Z)
C
C      ALGORITHM AS 197.1 APPLIED STATISTICS (1984) VOL.33, NO. 1
C
C      IMPLEMENTATION OF THE ALGORITHM OF G. TUNNICLIFFE WILSON
C      (J. STATIST. COMPUT. SIMUL. 8, 1979, 301-309) FOR THE COMPUTATION
C      OF THE AUTOCOVARIANCE FUNCTION (ACF) OF AN ARMA PROCESS OF ORDER
C      (MP, MQ) AND UNIT INNOVATION VARIANCE.   THE AUTOREGRESSIVE AND
C      MOVING AVERAGE COEFFICIENTS ARE STORED IN VECTORS P AND Q, USING
C      BOX AND JENKINS NOTATION.   ON OUTPUT, VECTOR CVLI CONTAINS THE
C      COVARIANCES BETWEEN THE VARIABLE AND THE (K-1)-LAGGED INNOVATION
C      FOR K = 1, ..., MQ+1.
C
C     ORIGINAL CODE FROM STATLIB USED P(MP),Q(MQ) BELOW, BUT THIS
C     FAILS ON COMPUTERS LIKE VAX/VMS WHEN MP OR MQ ARE ZERO.
      DIMENSION P(MP), Q(MQ), ACF(MA), CVLI(MXPQP1), ALPHA(MXPQ)
C
      DATA EPSIL2 /1.D-10/
      DATA ZERO, HALF, ONE, TWO /0.D0, 0.5D0, 1.D0, 2.D0/
C
      IFAULT = 0
      IF (MP .LT. 0 .OR. MQ .LT. 0) IFAULT = 1
      IF (MXPQ .NE. MAX0(MP, MQ)) IFAULT = 2
      IF (MXPQP1 .NE. MXPQ + 1) IFAULT = 3
      IF (MA .LT. MXPQP1) IFAULT = 4
      IF (IFAULT .GT. 0) RETURN
C
C      INITIALIZATION, AND RETURN IF MP = MQ = 0
C
      ACF(1) = ONE
      CVLI(1) = ONE
      IF (MA .EQ. 1) RETURN
      DO 10 I = 2, MA
   10 ACF(I) = ZERO
      IF (MXPQP1 .EQ. 1) RETURN
      DO 20 I = 2, MXPQP1
   20 CVLI(I) = ZERO
      DO 90 K = 1, MXPQ
   90 ALPHA(K) = ZERO
C
C      COMPUTATION OF THE A.C.F. OF THE MOVING AVERAGE PART,
C      STORED IN ACF.
C
      IF (MQ .EQ. 0) GO TO 180
      DO 130 K = 1, MQ
        CVLI(K+1) = -Q(K)
        ACF(K+1) = -Q(K)
        KC = MQ - K
        IF (KC .EQ. 0) GO TO 120
        DO 110 J = 1, KC
          JPK = J + K
          ACF(K+1) = ACF(K+1) + Q(J) * Q(JPK)
  110   CONTINUE
  120   ACF(1) = ACF(1) + Q(K) * Q(K)
  130 CONTINUE
C
C      INITIALIZATION OF CVLI = T.W.-S PHI -- RETURN IF MP = 0.
C
  180 IF (MP .EQ. 0) RETURN
      DO 190 K = 1, MP
        ALPHA(K) = P(K)
        CVLI(K) = P(K)
  190 CONTINUE
C
C      COMPUTATION OF T.W.-S ALPHA AND DELTA (DELTA STORED IN ACF
C      WHICH IS GRADUALLY OVERWRITTEN).
C
      DO 290 K = 1, MXPQ
        KC = MXPQ - K
        IF (KC .GE. MP) GO TO 240
        DIV = ONE - ALPHA(KC+1) * ALPHA(KC+1)
        IF (DIV .LE. EPSIL2) GO TO 700
        IF (KC .EQ. 0) GO TO 290
        DO 230 J = 1, KC
          KCP1MJ = KC + 1 - J
          ALPHA(J) = (CVLI(J) + ALPHA(KC+1) * CVLI(KCP1MJ)) / DIV
  230   CONTINUE
  240   IF (KC .GE. MQ) GO TO 260
        J1 = MAX0(KC + 1 - MP, 1)
        DO 250 J = J1, KC
          KCP1MJ = KC + 1 - J
C          ORIGINAL CODE FROM STATLIB
C          ACF(J+1) = ACF(J+1) + ACF(K+2) * ALPHA(KCP1MJ)
C          FIXED TO AGREE WITH JOURNAL LISTING
          ACF(J+1) = ACF(J+1) + ACF(KC+2) * ALPHA(KCP1MJ)
  250   CONTINUE
  260   IF (KC .GE. MP) GO TO 290
        DO 270 J = 1, KC
  270   CVLI(J) = ALPHA(J)
  290 CONTINUE
C
C      COMPUTATION OF T.W.-S NU (NU IS STORED IN CVLI, COPIED INTO
C      ACF).
C
      ACF(1) = HALF * ACF(1)
      DO 330 K = 1, MXPQ
        IF (K .GT. MP) GO TO 330
        KP1 = K + 1
        DIV = ONE - ALPHA(K) * ALPHA(K)
        DO 310 J = 1, KP1
          KP2MJ = K + 2 - J
          CVLI(J) = (ACF(J) + ALPHA(K) * ACF(KP2MJ)) / DIV
  310   CONTINUE
        DO 320 J = 1, KP1
  320   ACF(J) = CVLI(J)
  330 CONTINUE
C
C      COMPUTATION OF ACF (ACF IS GRADUALLY OVERWRITTEN).
C
      DO 430 I = 1, MA
        MIIM1P = MIN0(I-1, MP)
        IF (MIIM1P .EQ. 0) GO TO 430
        DO 420 J = 1, MIIM1P
          IMJ = I - J
          ACF(I) = ACF(I) + P(J) * ACF(IMJ)
  420   CONTINUE
  430 CONTINUE
      ACF(1) = ACF(1) * TWO
C
C      COMPUTATION OF CVLI - RETURN WHEN MQ = 0.
C
      CVLI(1) = ONE
      IF (MQ .LE. 0) GO TO 600
      DO 530 K = 1, MQ
        CVLI(K+1) = -Q(K)
        IF (MP .EQ. 0) GO TO 530
        MIKP = MIN0(K, MP)
        DO 520 J = 1, MIKP
          KP1MJ = K + 1 - J
          CVLI(K+1) = CVLI(K+1) + P(J) * CVLI(KP1MJ)
  520   CONTINUE
  530 CONTINUE

  600 RETURN
C
C      EXECUTION ERROR DUE TO (NEAR) NON-STATIONARITY.
C
  700 IFAULT = 5
      RETURN
      END





